/*
	Sherlock delete program.

	source:  sdel.c
	started: December 15, 1985
	version: April 2, 1996.
*/

#include "sdel.h"
#include "sdel_arg.h"
#include "sdel_init.h"
#include <LIBend.h>
#include <mac_gui.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

// Public globals.

char * in_file_name = NULL;		/* Input file name. */
char * out_file_name = NULL;	/* Output file name. */
char * syn_file_name = NULL;	/* Synonym file name. */

FILE * in_file = NULL;	/* Input file. */
FILE * out_file = NULL;	/* Output file. */
FILE * syn_file = NULL;	/* Synonym file. */

long	line = 1;	/* Current line number. */
short	ch = ' ';	/* The NEXT character. */

char *	r_void    = "RETURN_VOID";	/* Default spellings. */
char *	s_disable = "SL_DISABLE";
char *	s_name    = "SL_NAME";

/*
	Default spellings of all return macros, except RETURN_VOID.
*/
char * rmn_tab [] = {

	"RETURN_BOOL",	"RETURN_CHAR",	 "RETURN_DOUBLE",
	"RETURN_FLOAT",	"RETURN_INT",	 "RETURN_LONG",
	"RETURN_PTR",	"RETURN_STRING", "RETURN_UINT",
	"RETURN_ULONG",

	/* End of table flag. */
	NULL
};

/*
	Default spellings of all other macros.
*/
char * mn_tab [] = {

	"FTAG",

	"SL_CALL_NAME",	"SL_CLEAR",
	"SL_DIALOG",
	"SL_DUMP",		"SL_ENABLE",
	"SL_INIT",		"SL_MAC_INIT",
	"SL_OFF",		"SL_ON",
	"SL_PARSE",		"SL_SETNAME",
	"SL_UPDATE",

	"STAT",		"STATB",	"STATX",
	"TICK",		"TICKB",	"TICKN",
	
	"TICKX",
	
	"TICKX_BOOL",	"TICKX_CHAR",
	"TICKX_DOUBLE",	"TICKX_FLOAT",
	"TICKX_INT",	"TICKX_LONG",
	"TICKX_PTR",	"TICKX_STRING",
	"TICKX_UINT",	"TICKX_ULONG",
	"TICKX_VOID",
	
	"TRACE",	"TRACEB",	"TRACEN",	"TRACEX",
	"TRACEP",	"TRACEPB",	"TRACEPN",	"TRACEPX",

	NULL
};


// Private globals

static short last_ch	= '@';		/* Previous non-blank char.	*/
static bool	skipping	= FALSE;	/* TRUE if skipping output.	*/
static bool	delete_flag	= FALSE;	/* Newline deleted flag.	*/

#define MAX_SYMBOL 1000

static char	sym [MAX_SYMBOL];
static int	held_nl;	/* Number of held newlines.	*/

static char lws [MAX_SYMBOL];	/* Leading white space.		*/
static int	lws_count = 0;

static char tws [MAX_SYMBOL];	/* Trailing white space.	*/
static int	tws_count = 0;

// Private prototypes

static void	delete		(void);
static void	do_comment	(void);
static void	do_id		(bool nl_flag, bool multiple_flag);
static void	do_pp		(void);
static void	do_single	(void);
static void	do_string	(void);
static void	do_ws		(void);
static bool	is_return	(char * name);
static bool	is_sherlock	(char * name);
static void	out_all		(void);
static void out_char	(short c);
static void	out_hnl		(void);
static void	out_lws		(void);

/*
	Main routine.
*/
void
main(int argc, char ** argv)
{
	FTAG("main");
	
	init_all();
	
	STATB(ftag);

	
	delete();
	main_quit();
	
	STATX(ftag);
}

void
main_quit (void)
{
	if (in_file)
		fclose(in_file);
	if (out_file)
		fclose(out_file);
	if (syn_file)
		fclose(syn_file);
	end_done();
}

/*
	Delete all calls to Sherlock macros, with any leading white space.
*/
static void
delete(void)
{
	FTAG("delete");

	int eventCount = 1;

	STATB(ftag);

	if (ch == '#') {
		do_pp();
	}
	
	while(ch != EOF) {
	
		if (!w_applEvent(10)) {
			ecnl(); es("SDEL terminated by user.\n");
			main_quit();
		}

		if (delete_flag) {
			delete_flag = FALSE;
			held_nl = 0;

			if (isalpha(ch) || ch == '_') {
				/* Leading id, multiple delete possible */
				do_id(TRUE, TRUE);
			}
			else if (ch == '#') {
				/* Allow preprocessor directives. */
				out_lws();
				do_pp();
			}
			else {
				out_lws();
			}
		}
		else if (ch == '\n') {

			/*
				Count the number of newline and queue up
				the last leading white space.
			*/
			skipping = TRUE;
			
			held_nl = 0;
			for (;;) {
				if (ch == '\n') {
					line++;
					held_nl++;
					next_ch();
					lws_count = 0;
				}
				else if (ch == ' ' || ch == '\t') {
					lws[lws_count++] = ch;
					next_ch();
				}
				else {
					break;
				}
			}
			skipping = FALSE;
					
			if (isalpha(ch) || ch == '_') {
				/* Leading id, no multiple delete */
				do_id(TRUE, FALSE);
			}
			else if (ch == '#') {
				out_hnl();

				/* Allow preprocessor directives. */
				do_pp();
			}
			else {
				out_all();
			}
		}
		else if (isalpha(ch) || ch == '_') {
			/* non-leading id, no multiple line. */
			do_id(FALSE, FALSE);
			last_ch = 'a';
		}
		else if (ch == '"' || ch == '\'') {
			do_string();
			last_ch = '"';
		}
		else if (ch == '/') {
			out_char(ch);
			next_ch();
			if (ch == '*') {
				out_char(ch);
				next_ch();
				do_comment();
			}
			else if (ch == '/') {
				out_char(ch);
				next_ch();
				do_single();
			}
			else {
				last_ch = '/';
			}
		}
		else if (ch == '*') {
			out_char(ch);
			next_ch();
			if (ch == '/') {
				warning(es("'*/' seen outside a comment."));
				out_char(ch);
				next_ch();
			}
			else {
				last_ch = '*';
			}
		}
		else {
			STAT("v_default");
			last_ch = ch;
			out_char(ch);
			next_ch();
		}
	}
	
	STATX(ftag);
}

/*
	Handle all remaining actual arguments to a Sherlock macro.
	The macro should end with a right paren and a semicolon.
*/
static void
do_arg(void)
{
	FTAG("do_arg");
	
	int level;
	int start;

	STATB("do_arg");
	TRACE("v", os("<BEGIN ARG>"));

	level = 1;
	start = line;

	for (;;) {
		do_ws();
		if (ch == EOF) {
			error(
				es("starting line: "); eint(start);
				es(": Unexpected End of File inside a Sherlock macro"));
			
			STATX(ftag);
			return;
		}
		else if (ch == '(') {
			level++;
			out_char(ch);
			next_ch();
		}
		else if (ch == ')') {
			level--;
			if (level == 0) {
				TRACE("v", os("<END ARG>"));
				STATX(ftag);
				return;
			}
			else {
				out_char(ch);
				next_ch();
			}
		}
		else if (ch == '\n') {
			line++;
			if (!skipping) {
				onl();
			}
			next_ch();
		}
		else if (ch == '"' || ch == '\'') {
			do_string();
		}
		else {
			out_char(ch);
			next_ch();
		}
	}
}

/*
	Handle a comment, including the closing, but not opening delim.
*/
static void
do_comment(void)
{
	FTAG("do_comment");

	int level;
	int start;

	STATB(ftag);
	TRACE("v", os("<BEGIN COMMENT>"));

	start = line;
	level = 1;

	for (;;) {
		if (ch == EOF) {
			error(
				es("starting line: "); eint(start);
				es(": Unexpected End of File inside a comment"));
			
			STATX(ftag);
			return;
		}
		else if (ch == '/') {
			out_char(ch);
			next_ch();
			if (ch == '*') {
				out_char(ch);
				next_ch();
				if (nest_flag) {
					level++;
				}
				else {
					warning(es("'/*' in comment ignored."));
				}
			}
		}
		else if (ch == '*') {
			out_char(ch);
			next_ch();
			if (ch == '/') {
				out_char(ch);
				next_ch();
				level--;
				if (level == 0) {
					TRACE("v", os("<END COMMENT>"));
					STATX(ftag);
					return;
				}
			}
		}
		else if (ch == '\n') {
			line++;
			if (!skipping) {
				onl();
			}
			next_ch();
		}
		else {
			out_char(ch);
			next_ch();
		}
	}
}

/*
	Handle an identifier.
	nl_flag:  True if the id is the first non-white space of the line.
*/
static void
do_id (bool nl_flag, bool multiple_flag)
{
	FTAG("do_id");

	int start;
	int i;

	STATB("do_id");

	/* Get the id into sym[]. */
	get_id(sym);

	if (str_eq(sym, r_void)) {
		start = line;

		/* Output all queued newlines and leading white space. */
		out_all();

		/* Translate RETURN_VOID to return */
		os("return");

		skipping = TRUE;

		do_ws();
		if (ch == '(') {
			next_ch();
		}
		else {
			error(es("'(' expected after Sherlock macro."));
		}
		
		do_arg();
		if (ch == ')') {
			/* Skip the trailing closing paren. */
			next_ch();
		}
		do_ws();

		skipping = FALSE;

		if (ch == ';') {
			out_char(ch);
			next_ch();
		}
		else {
			error(
				es("starting line: "); eint(start);
				es(": Semicolon expected after Sherlock macro"));
		}
	}
	else if (is_return(sym)) {
		start = line;

		/* Output queued newlines and leading white space. */
		out_all();

		/* Translate RETURN_xxx to return */
		os("return");

		do_ws();
		if (ch == '(') {
			next_ch();
		}
		else {
			error(es("'(' expected after Sherlock macro."));
		}
		
		skipping = TRUE;

		/* We can't demand a string here.  An array is OK too. */
		do_ws();

		while (ch != ',' && ch != EOF) {
			next_ch();
		}
		if (ch == ',') {
			next_ch();
		}
		else {
			error(es("Unexpected End of File in Sherlock macro."));
			STATX(ftag);
			return;
		}

		/* Copy the argument. */
		skipping = FALSE;

		do_arg();
		if (ch == ')') {
			/* Skip the trailing closing paren. */
			next_ch();
		}
		do_ws();
		if (ch == ';') {
			out_char(ch);
			next_ch();
		}
		else {
			error(
				es("starting line: "); eint(start);
				es(": Semicolon expected after Sherlock macro"));
		}
	}
	else if (is_sherlock(sym)) {
		start = line;

		/* Throw away all saved white space. */
		if (nl_flag) {
			held_nl = 0;
			lws_count = 0;
		}
		else {
			out_all();
		}
		skipping = TRUE;

		do_ws();
		if (ch == '(') {
			next_ch();
		}
		else {
			error(es("'(' expected after Sherlock macro."));
		}
		do_arg();
		if (ch == ')') {
			next_ch();
		}
		if (ch == ';') {
			next_ch();
		}
		else {
			error(
				es("starting line: "); eint(start);
				es(": Semicolon expected after Sherlock macro"));
		}

		skipping = FALSE;

		if (!nl_flag) {
			STATX(ftag);
			return;
		}

		/* Save trailing white space. */
		while (ch == ' ' || ch == '\t') {
			tws [tws_count++] = ch;
			next_ch();
		}

		if (ch != '\n') {
			/* Change everything to one newline. */
			onl();
			for (i = 0; i < tws_count; i++) {
				out_char(tws[i]);
			}
			STATX(ftag);
			return;
		}

		/* Discard trailing white space. */
		tws_count = 0;

		/* Indicate at least one newline was deleted. */
		delete_flag = TRUE;

		/* Discard all newlines and all leading white space. */
		while (ch == '\n') {
			next_ch();
			lws_count = 0;
			while (ch == ' ' || ch == '\t') {
				lws[lws_count++] = ch;
				next_ch();
			}
		}

		/*
			Let the main line handle the newline so that
			leading white space will be handled properly.
		*/
		if (!multiple_flag) {
			if (ch != '}' && last_ch != '{') {
				onl();
				onl();
			}
			else {
				onl();
			}
		}
		skipping = FALSE;
	}
	else {
		out_all();
		os(sym);
	}
	skipping = FALSE;
	STATX(ftag);
}

/*
	Handle a preprocessor directive.
*/
static void
do_pp(void)
{
	FTAG("do_pp");

	int old_skipping;

	STATB(ftag);
	TRACE("v", os("<BEGIN PP>"));

	old_skipping = skipping;
	skipping = FALSE;

	for (;;) {
		do_ws();
		if (ch == EOF) {
			skipping = old_skipping;

			TRACE("v", os("<END PP>"));
			STATX(ftag);
			return;
		}
		else if (ch == '\\') {
			out_char(ch);
			next_ch();
			if (ch == '\n') {
				line++;
				onl();
				next_ch();
			}
		}
		else if (ch == '\n') {
			skipping = old_skipping;

			/*
				Let the main line handle the newline so
				consecutive directives will be recognized.
			*/
			TRACE("v", os("<END PP>"));
			STATX(ftag);
			return;
		}
		else if (ch == '"' || ch == '\'') {
			do_string();
		}
		else {
			out_char(ch);
			next_ch();
		}
	}
}

/*
	Copy a single-line comment.
*/
static void
do_single(void)
{
	while(ch != '\n' && ch != EOF) {
		out_char(ch);
		next_ch();
	}
}

/*
	Handle a string, including the opening and closing delimiters.
*/
static void
do_string(void)
{
	FTAG("do_string");

	int delim;
	int start;
	int i;

	STATB(ftag);
	TRACE("v", os("<START STRING>"));

	start = line;
	delim = ch;

	out_char(ch);
	next_ch();

	for (;;) {
		if (ch == delim) {
			out_char(ch);
			next_ch();

			TRACE("v", os("<END STRING>"));
			STATX(ftag);
			return;
		}
		else if (ch == EOF) {

			if (ch == '"') {
				error(
					es("starting line: "); eint(start);
					es(": Unexpected End of File inside string "));
			}
			else {
				error(
					es("starting line: "); eint(start);
					es(": Unexpected End of File inside character constant "));
			}
			STATX(ftag);
			return;
		}

		else if (ch == '\n') {
			line++;
			next_ch();
			if (!skipping) {
				onl();
			}
			error(
					es("starting line: "); eint(start);
					es(": Run on string: newline in string "));
			
			STATX(ftag);
			return;
		}
		else if (ch == '\\') {
			/* Escape sequence. */

			STAT("v_esc");

			out_char(ch);
			next_ch();
			if (ch == '\n') {
				/* Strings eat directives. */
				line++;
				next_ch();
				if (!skipping) {
					onl();
				}
			}
			else if (ch == 'x') {
				/* 3-digit hex escape sequence. */
				out_char(ch);
				next_ch();
				for (i = 0; i < 3 && ishex(ch); i++) {
					out_char(ch);
					next_ch();
				}
			}
			else if (ch >= '0' && ch <= '9') {
				/* Octal escape sequence. */
				for (i = 0; i < 3 && ch >= '0' && ch <= '9'; i++) {
					out_char(ch);
					next_ch();
				}
			}
			else if (ch != EOF) {
				out_char(ch);
				next_ch();
			}
		}
		else {
			out_char(ch);
			next_ch();
		}
	}
}

/*
	Handle white space, including comments.
*/
static void
do_ws(void)
{
	FTAG("do_ws");
	STATB(ftag);

	for (;;) {
		if (ch == ' ' || ch == '\t') {
			out_char(ch);
			next_ch();
		}
		else if (ch == '/') {
			out_char(ch);
			next_ch();
			if (ch == '*') {
				do_comment();
			}
		}
		else {
			STATX(ftag);
			return;
		}
	}
}

/*
	Get an identifier from the input file into the buffer.
*/
void
get_id(char * buffer)
{
	int i = 0;

	while (isid2(ch)) {
		buffer[i++] = ch;
		next_ch();
	}
	buffer[i] = '\0';
}
		
/*
	Return TRUE if name is a Sherlock RETURN_xxx macro except RETURN_VOID.
*/
static bool
is_return(char * name)
{
	int i;
	for (i = 0; rmn_tab[i]; i++)
		if (str_eq(rmn_tab[i], name))
			return TRUE;

	return FALSE;
}

/*
	Return TRUE if name is a Sherlock macro except RETURN_xxx.
*/
static bool
is_sherlock(char * name)
{
	int i;

	/* Special cases, but not r_void. */
	if (str_eq(s_disable, name))
		return !disable_flag;
	else if (str_eq(s_name, name))
		return !name_flag;

	/* All other macros. */
	for (i = 0;;i++) {
		if (mn_tab[i] == NULL) {
			return FALSE;
		}
		else if (str_eq(mn_tab[i], name)) {
			return TRUE;
		}
	}
}

/*
	Output all queued white space and newlines.
*/
static void
out_all(void)
{
	out_hnl();
	out_lws();
}

/*
	Output a single character.
	Perform trigraph translations.
*/
static void
out_char(short c)
{	
	if (!skipping) {
		if (trigraph_flag) {
			switch (c) {
			case '#':	os("??=");	break;
			case '[':	os("??(");	break;
			case '\\':	os("??/");	break;
			case ']':	os("??)");	break;
			case '^':	os("??'");	break;
			case '{':	os("??<");	break;
			case '|':	os("??!");	break;
			case '}':	os("??>");	break;
			case '~':	os("??-");	break;
			default:
				ochar(c);
			}
		}
		else {
			ochar(c);
		}
	}
}

/*
	Output held newlines.
*/
static void
out_hnl(void)
{
	int i;
	for (i = 0; i < held_nl; i++)
		onl();
	held_nl = 0;
}

/*
	Output leading white space.
*/
static void
out_lws(void)
{
	int i;
	for (i = 0; i < lws_count; i++)
		out_char(lws [i]);
	lws_count = 0;
}

/*
	Skip blanks and tabs.
*/
void
skip_bl(void)
{
	while (ch == ' ' || ch == '\t')
		next_ch();
}
