/*
	SDEL: routines that return the next character.
	
	source:  sdel_next.c
	started: April 2, 1996.
	version: April 2, 1996.
*/

#include "sdel.h"
#include "sdel_arg.h"

#include <string.h>
#include <stdio.h>

// Private data

char *	s_ip = NULL;    /* input pointer. */

#define	MAX_LINE 1000
char 	s_buf [MAX_LINE];	/* Line buffer. */
char *	s_bufp = s_buf;		/* Line buffer pointer. */

// Private prototypes

static void	fill_buf	(void);
static int	fill1		(void);

/*
	Fill the line buffer.
	Delete lines consisting of #include "sl.h" or <sl.h>.
	Set buffer[0] to '\0' if the buffer must be refilled.
*/
static void
fill_buf(void)
{
	int c;
	
	s_bufp = s_buf;
	do
	{
		*s_bufp++ = c = fill1();
	}
	while (c != EOF && c != '\n');

	*s_bufp = '\0';
	s_bufp = s_buf;

	if (
		include_flag &&
		(
			str_eq(s_bufp, "#include \"sl.h\"\n") ||
			str_eq(s_bufp, "#include <sl.h>\n")
		)
	) {
		line++;
		s_buf[0] = '\0';
	}
}

/*
	Return the next character to fill_buf().
	Translate trigraphs.
*/
static int
fill1(void)
{
	int c = fgetc(in_file);

	if (trigraph_flag && c == '?')
	{
		c = fgetc(in_file);
		if (c == '?')
		{
			/* A trigraph sequence. */
			c = fgetc(in_file);
			switch (c) {
			case '=':  c = '#';	 break;
			case '(':  c = '[';	 break;
			case '/':  c = '\\'; break;
			case ')':  c = ']';	 break;
			case '\'': c = '^';	 break;
			case '<':  c = '{';	 break;
			case '!':  c = '|';	 break;
			case '>':  c = '}';	 break;
			case '-':  c = '~';	 break;
			default:
				es("Warning: line "); eint(line);
				es(": unknown trigraph: ??"); echar(c);
				es(" ignored\n");
			}
		}
		else
		{
			ungetc(c, in_file);
			c = '?';
		}
	}
	return c;
}
	
/*
	Set global variable ch to the next character from the input stream.
*/
void
next_ch(void)
{
	FTAG("next_ch");
	STATB(ftag);
	ASSERT(in_file);
	
	while (*s_bufp == '\0')
		fill_buf();
		
	ch = *s_bufp;

	if (ch != EOF)
		++s_bufp;
	
	STATX(ftag);
}
