/*
	Sherlock File Comparison Program.

	source:  sdif.c
	started: February 11, 1988
	version: April 2, 1996.
*/

#include "sdif.h"
#include "sdif_arg.h"
#include "sdif_init.h"

#include <LIBend.h>
#include <mac_gui.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>


/* Public globals: the windows... */

w_type w1, w2;

/*
	Function prototypes of the routines in this file.
*/
static void	advance			(w_type *wp);
static bool	compare			(int n1, int n2);
static void	do_argv			(int arg, char ** argv);
static void	err_abort		(char * message);
static bool	fill_buf		(char * buffer, w_type *wp);
static void	insert			(w_type *wp);
static void	print_change	(int n);
static void	print_insert	(int n1, int n2);
static void	print_match		(void);
static bool	resynch			(int n1, int n2);
static void	sdif			(void);

/*
	Main routine.  Process command line arguments.
*/
int
main(int argc, char **argv)
{
	SL_DISABLE();
	
	init_all();
	sdif();
	
	end_done();
	main_quit();
}

void
main_quit (void)
{
	if (w1.file)
		fclose(w1.file);
	if (w2.file)
		fclose(w2.file);

	end_exit();
}

/*
	Advance one line in the indicated window.
	This frees up space at the beginning of the window buffer.
*/
static void
advance(w_type *wp)
{
	FTAG("advance");

	int freed;
	int i;
	int lines;

	TRACEPB(ftag, eptr(wp); enl());

	lines = wp -> nlines;
	if (lines == 0) {
		STATX(ftag);
		return;
	}
	freed = strlen(&wp->window[wp->index[0]])+1;

	wp -> first += freed;
	lines--;
	for (i = 0; i < lines; i++) {
		wp -> index[i] = wp -> index[i+1];
	}
	wp -> nlines--;
	wp -> line++;

	/* Refill the buffer. */
	if (wp -> nlines == WINDOW_LINES-1) {
		insert(wp);
	}

	TICKX(ftag);
}

/*
	Return TRUE if the indicated lines match.
*/
static bool
compare(int n1, int n2)
{
	FTAG("compare");
	
	char *p1, *p2;
	bool result = 0;

	STATB(ftag);
	
	if (n1 >= w1.nlines || n2 >= w2.nlines) {
		result = FALSE;
	}
	else {
		p1 = &w1.window[w1.index[n1]];
		p2 = &w2.window[w2.index[n2]];
		result = str_eq(p1, p2);
	}
	
	TRACEPX(ftag,
		eint(n1); ecs(); eint(n2); eblank();
		eret(); ebool(result); enl());

	return result;
}

/*
	Fill a buffer from a file.
	Set the end of file flag if appropriate.
*/
static bool
fill_buf(char *buffer, w_type *wp)
{
	FTAG("fill_buf");
	
	int c;
	int i;
	bool result;
	
	STATB(ftag);

	if (wp -> eof) {
		result = FALSE;
		goto done;
	}

	for (i = 0;;) {
	
		c = fgetc(wp -> file);

		if (c == '\r') {
			continue;
		}
		if (c == EOF) {
			wp -> eof = TRUE;
			if (i == 0) {
				result = FALSE;
				goto done;
			}
			break;
		}
		else if (c == '\n') {
			buffer[i++] = c;
			break;
		}
		else {
			buffer[i++] = c;
		}
	}

	buffer[i] = '\0';
	result = TRUE;
	
done:
	TRACEPX(ftag,
		eptr(buffer); ecs(); eptr(wp); eblank();
		eret(); ebool(result); enl());

	return result;
}

/*
	Insert a line at the end of the window.
	Pack the buffer if required.
*/
static void
insert(w_type *wp)
{
	FTAG("insert");
	
	char buffer [1000];
	int size, avail;
	int i, p, q;

	TRACEPB(ftag, eptr(wp); enl());
	
	if (!fill_buf(buffer, wp)) {
		STATX(ftag);
		return;
	}
	size = strlen(buffer)+1;
	avail = WINDOW_CHARS - wp -> last;

	if (wp -> nlines >= WINDOW_LINES) {
		es("insert: too many lines.\n");
		STATX(ftag);
		return;
	}

	/* Compact buffer. */
	if (size >= avail) {
		/* Adjust indices. */
		for (i = 0; i < wp -> nlines; i++) {
			wp -> index[i] -= wp -> first;
		}
		/* Move the characters in the buffer. */
		for (p = wp -> first, q = 0; p < wp -> last; p++, q++) {
			wp -> window[q] = wp -> window[p];
		}
		/* Adjust counts. */
		wp -> last -= wp -> first;
		avail      += wp -> first;
		wp -> first = 0;
	}

	/* Insert the buffer at the end of the window. */
	if (size < avail) {
		strcpy(&wp -> window[wp -> last], buffer);
		wp -> index[wp -> nlines] = wp -> last;
		wp -> last += size;
		wp -> nlines++;
	}
	else {
		es("not enough room in window!!\n");
		main_quit();
	}

	STATX(ftag);
}

/*
	Print a changed line.
*/
static void
print_change(int n)
{
	FTAG("print_change");

	char * p1 = &w1.window[w1.index[n]];
	char * p2 = &w2.window[w2.index[n]];
	char * p = p1;

	TRACEPB(ftag, eint(n); enl());

	/* Do not print mismatches that involve only white space. */
	if (!s_flag && !v_flag && !b_flag) {
		while (*p1) {
			if (*p1 != ' ' && *p1 != '\t' && *p1 != '\n') {
				goto print;
			}
			p1++;
		}
		while (*p2) {
			if (*p2 != ' ' && *p2 != '\t' && *p2 != '\n') {
				goto print;
			}
			p2++;
		}
		STATX(ftag);
		return;
	}

print:
	epadint(w1.line+n, 4); eblanks(5);
	es(v_flag ? "*" : ":"); eblank();
	es(p1); ecnl();
	
	eblanks(5); epadint(w2.line+n, 4);
	es(v_flag ? "*" : ":"); eblank();
	es(p2); ecnl();

	STATX(ftag);
}

/*
	Print an inserted line.
	The line comes from w1.name if n1 >0, or from n2 if n2 > 0.
*/
static void
print_insert(int n1, int n2)
{
	FTAG("print_insert");
	
	char *p;

	TRACEPB(ftag, eint(n1); ecs(); eint(n2); enl());

	/* Do not print blank lines. */
	if (!b_flag && !v_flag && n1 >= 0) {
		p = &w1.window[w1.index[n1]];
		while (*p) {
			if (*p != ' ' && *p != '\t' && *p != '\n') {
				goto print;
			}
			p++;
		}
		STATX(ftag);
		return;
	}
	else if (!b_flag && !v_flag && n2 >= 0) {
		p = &w2.window[w2.index[n2]];
		while (*p) {
			if (*p != ' ' && *p != '\t' && *p != '\n') {
				goto print;
			}
			p++;
		}
		STATX(ftag);
		return;
	}

print:
	if (n1 >= 0) {
		epadint(w1.line+n1, 4);
		eblanks(5); es(": ");
		es(&w1.window[w1.index[n1]]);
	}
	else {
		epadint(w2.line+n1, 4);
		eblanks(5); es(": ");
		es(&w2.window[w2.index[n2]]);
	}
	
	STATX(ftag);
}

/*
	Print a matched line if the -v option was given.
*/
static void
print_match(void)
{
	FTAG("print_match");
	STATB(ftag);

	if (v_flag) {
		epadint(w1.line, 4); eblank();
		epadint(w2.line, 4); es(": ");
		es(&w1.window[w1.index[0]]);
	}

	STATX(ftag);
}

/*
	Return TRUE if the indicated lines match and can be used to
	resynchronize the files.
*/
static bool
resynch(int n1, int n2)
{
	FTAG("resynch");

	char *p;
	int count;
	bool result;

	TRACEPB(ftag, eint(n1); ecs(); eint(n2); enl());

	if (!compare(n1, n2)) {
		result = FALSE;
		goto done;
	}

	p = &w1.window[w1.index[n1]];

	/* Make sure we have a non-trivial resynch point. */
	count = 0;
	while (*p) {
		if (*p != ' ' && *p != '\t' && *p != '\n') {
			count++;
		}
		p++;
	}

	if (
		count >= 1 ||
		(w1.eof && w2.eof && n1 == w1.nlines-1 && n2 == w2.nlines-1)
	) {
		/* Non-trivial matched lines or match to end of file. */
		result = TRUE;
	}
	else {
		/* Trivial matched lines.  Look ahead for an answer. */
		result = resynch(n1+1, n2+1);
	}
	
done:
	TRACEPX(ftag, eret(); ebool(result); enl());
	return result;
}

/*
	Compare two files line by line.
	Print lines that do not match.
	
	Assume that file 1 contains any inserted lines.
	This assumption greatly simplifies the general problem
	and allows a simple and fast solution.
*/
static void
sdif(void)
{
	FTAG("sdif");

	int i, j;

	STATB(ftag);

	/* Fill up the window buffers. */
	for (i = 0; i < WINDOW_LINES; i++) {
		insert(&w1);
		insert(&w2);
	}

loop:

	if (!w_applEvent(10)) {
		es("SDIF terminated by user.\n");
		main_quit();
	}

	if (w1.nlines == 0 && w2.nlines == 0) {
		STATX(ftag);
		return;
	}
	else if (w1.nlines == 0 && w2.nlines >= 10) {
		ecnls(2);
		es("File "); es(w1.name);
		es(" ends before file "); es(w2.name); enl();
		STATX(ftag);
		return;
	}
	else if (w2.nlines == 0 && w1.nlines >= 10) {
		ecnls(2);
		es("File "); es(w2.name);
		es(" ends before file "); es(w1.name); enl();
		STATX(ftag);
		return;
	}

	if(compare(0, 0)) {
		/* Lines match. */
		print_match();
		advance(&w1);
		advance(&w2);
		goto loop;
	}

	/* Look for some changed or inserted lines. */
	for (i = 1; i < WINDOW_LINES; i++) {
		/* Don't resynch on duplicated lines. */
		if (resynch(i, i) && !compare(i, i+1)) {
			if (s_flag) {
				es("----- "); eint (i);
				es(" changed line");
				if (i > 1) es("s");
				enl();
			}
			for (j = 0; j < i; j++) {
				print_change(j);
			}
			for (j = 0; j < i; j++) {
				advance(&w1);
				advance(&w2);
			}
			goto loop;
		}

		if (resynch(i, 0)) {
			if (s_flag) {
				es("----- "); eint(i);
				es(" inserted line");
				if (i > 1) es("s");
				enl();
			}
			for (j = 0; j < i; j++) {
				print_insert(j, -1);
			}
			for (j = 0; j < i; j++) {
				advance(&w1);
			}
			goto loop;
		}
	}

	/*
		Look for lines inserted in file 2.
		This can happen as a result of previous erroneous advances.
	*/
	for (i = 1; i < WINDOW_LINES; i++) {
		if (resynch(0, i)) {
			if (s_flag) {
				es("----- "); eint(i);
				es(" back inserted line");
				if (i > 1) es("s");
				enl();
			}
			for (j = 0; j < i; j++) {
				print_insert(-1, j);
			}
			for (j = 0; j < i; j++) {
				advance(&w2);
			}
			goto loop;
		}
	}

	/*
		We haven't identified either a single group of insertions or
		a single group of changed lines.  We have probably just seen
		a combination of changes and insertions.  Just advance both
		files one line each.  We'll get back in synch quickly.
	*/

	if (s_flag) {
		es("----- failure advance\n");
	}
	if (w1.nlines) {
		print_insert(0, -1);
	}
	else {
		print_insert(-1, 0);
	}
	advance(&w1);
	advance(&w2);
	goto loop;
}
