/*
	SDIF -- Global header file.
	
	source:  sdif.h
	started: June 8, 1991
	version: April 2, 1996.
*/

#ifndef sdif_h_
#define sdif_h_

#pragma once

#include <LIBlib.h>
#include <stdio.h>

/*
	Each file is associated with a window for comparing lines.

	Lines are inserted from the back of the window buffers and deleted from
	the front.  When the back of the window buffer can not hold the next 
	line, the non-deleted lines are moved to the front of the buffer.  The
	window buffer is made much larger than required	lines so this moving of
	lines in the window buffer doesn't happen often.

	We expect an average line size of less than 40, so that WINDOW_LINES
	lines will take about 160 characters.  Thus, the window will have to be 
	repacked about once in every 100 inserted lines.  This will not
	slow down the program in any way.
*/
#define WINDOW_LINES	20
#define WINDOW_CHARS	7000
#define BUF_SIZE		512

typedef struct {
	char *	name;					/* File name. */
	FILE *  file;					/* File handle for fopen. */
	bool	eof;					/* End of file flag. */
	int		line;					/* Current line number. */
	int		nlines;					/* # of lines in window. */
	int		index[WINDOW_LINES];	/* Indices into window. */
	char	window[WINDOW_CHARS];	/* Chars of window. */
	int		first;					/* Index of first character. */
	int		last;					/* Index of last character+1. */
} w_type;

extern w_type w1, w2;	/* The comparison windows. */

#endif /* sdif_h_ */

