/*
	SPP: Argument processing module
	
	source:  arg.c
	started: November 30, 1990
	version: March 17, 1996.
*/

#include "SPP.h"

#include "SPParg.h"
#include "SPPhold.h"
#include "SPPmst.h"
#include "SPPobj.h"
#include "SPPres.h"
#include "SPPnext.h"
#include "SPPtok.h"
#include "SPPver.h"

#include "SPPend.h"	/* Library additions. */
#include "SPPenv.h"
#include "SPPerr.h"
#include "SPPlst.h"
#include "SPPstr.h"

#include <stdlib.h>

/*
	Define global user flags;
*/
bool exxx_flag	 = TRUE;	/* TRUE: generate exxx() calls. */
bool ftag_flag	 = TRUE;	/* TRUE: generate FTAG macros calls. */
bool insert_flag = FALSE;	/* TRUE: insert #include "sl.h" */
bool nest_flag   = FALSE;	/* TRUE: comments can nest. */
bool usr_flag    = FALSE;	/* TRUE: use user output routines. */
bool usr2_flag   = FALSE;	/* TRUE: use user format routines. */
bool tick_flag   = FALSE;	/* TRUE: use count statistics only. */
bool slc_flag    = TRUE;	/* TRUE: allow single-line comments. */

/*
	The following flags are never changed in the current version.
*/
bool entry_flag  = TRUE;	/* TRUE: generate Sherlock macros on entry. */
bool exit_flag   = TRUE;	/* TRUE: generate Sherlock macros on exit. */
bool loop_flag   = TRUE;	/* TRUE: generate Sherlock macros in loops. */
bool return_flag = FALSE;	/* TRUE: generate RETURN_XXX macros. */

/*
	Define the global file paths and files.
*/
io_path *	arg_arg_path		= NULL;
io_path *	arg_in_path			= NULL;
io_path *	arg_out_path		= NULL;
io_path *	arg_syn_path		= NULL;

io_path *	arg_search_paths	= NULL;	/* Do not re-initialize.	*/
io_path *	arg_paths			= NULL;	/* Do not re-initialize.	*/

io_file *	arg_arg_file		= NULL; /* Do not re-initialize.	*/

io_strm *	arg_in_strm			= NULL;
io_strm *	arg_out_strm		= NULL;

/*
	Define dynamically allocated buffers.
	arg_def_buf and arg_undef buf are used arg_dialog.c.
*/
#define ARG_FILE_BUF_SIZE 2000

char * arg_def_buf			= NULL;	/* Do not re-initialize.	*/
char * arg_undef_buf		= NULL;	/* Do not re-initialize.	*/

static char * arg_file_buf	= NULL; /* Do not re-initialize.	*/

static int		arg_save_argc	= 0;	/* Saved argc.			*/
static char **	arg_save_argv	= NULL;	/* Saved argp.			*/

/*
	Declare internal routines.
*/
static void arg_synonyms(void);
static void arg_set_syn	(char *, char *);
static void arg_help 	(void);

/*
	Set up the program's arguments which are found in an argv vector.
	
	This routine may be re-entered.
*/
#define debug_arg_do_argv(a) a

void
arg_do_argv(int argc, char ** argv, char * on_str, char * off_str)
{
	FTAG("arg_do_argv");

	reg char	* arg;
	char		* in,  * out;
	char		* def;
	char		* name;
	io_path		* path_p;
	io_file		* file;
	
	/*
		Do not put Sherlock macros here: Sherlock has not been initialized!
	*/
	SL_DISABLE();
	
	debug_arg_do_argv(
		TRACEP(ftag,
			es("argc: "); eint(argc);
			es("argv: "); eptr(argv)));
			
	if (arg_paths != NULL) {
		mst_restore();
		arg_in_path = arg_paths;
		lst_remove(arg_paths, &arg_paths);
		arg_out_path == NULL;
		return;
	}
	
	if (argv == NULL) {
		argc = arg_save_argc;
		argv = arg_save_argv;
		mst_restore();
	}
	else {
	
		/* Make first test for correct command line. */
		ASSERT(argv);

		if (argc == 2 && str_eq(argv[1], "-?")) {
			es(VER_VERSION); ecnl();
			end_usage();
		}
		else if (argc < 3) {
			arg_help();
			end_usage();
		}
	}

	/* Indicate that no file arguments have been seen yet. */
	in  = NULL;
	out = NULL;

	/* Process all the arguments on the command line. */
	ASSERT(argv);
	argc--;
	argv++;
	while (argc-- > 0) {
	
		ASSERT(*argv);
		arg = *argv++;
		
		debug_arg_do_argv(TRACEP(ftag, es(arg); enl()));
		
		if (arg[0] == '@') {
			if (str_eq(arg+1, "goto")) {
				if (argc--) {

					/* Scan forward for the label. */
					char * label = *argv++;
					ecnl(); es("@goto "); es(label); enl();
					while(argc--) {
						arg = *argv++;
						if (str_eq(label, arg)) {
							break;
						}
					}
				}
				else {
					es("Trailing @goto\n");
					end_usage();
				}
			}
			else {
				/* Ignore the label. */
				;
			}
		}
		else if (arg[0] == '+' || arg[0] == '-')
		{
			bool arg_bool = (arg[0] == '+');
			arg++;

			switch(arg[0]) {
			
			case 'd':
			
				if(str_eq(arg, "d")) {
					/* Define a variable. */
					if (argc--) {
						arg = *argv++;
						/* Scan for an optional equal sign. */
						for (def = arg;	*def; def++) {
							if (*def == '=') {
								*def = '\0';
								def++;
								break;
							}
						}
						mst_enter(arg, def, -1);
					}
					else {
						es("Trailing -d\n");
						end_usage();
					}
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'e':
				
				if (str_eq(arg, "exxx")) {
					exxx_flag = arg_bool;
					tick_flag = FALSE;
					usr_flag  = FALSE;
					usr2_flag = FALSE;
				}
				else if (str_eq(arg, "exit")) {
					goto no_more_args;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'f':
			
				if (str_eq(arg, "f")) {
					if (argc--) {
						arg = *argv++;
						
						path_p	= io_new_path_struct(arg);
						file	= io_open(path_p, TRUE);
						if (file == NULL) {
							end_usage();
						}
						hold_flag = TRUE;
						arg_synonyms();
						io_close(file);
						hold_kill();
						ch = ' ';
						hold_flag = FALSE;
						hold_ch   = FALSE;
					}
					else {
						es("Trailing -f\n");
						end_usage();
					}
				}
				else if (str_eq(arg, "ftag")) {
					ftag_flag = arg_bool;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'i':
			
				if (str_eq(arg, "i")) {
					insert_flag = arg_bool;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'n':

				if (str_eq(arg, "n")) {
					nest_flag = TRUE;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'o':
			
				if (str_eq(arg, "of")) {
					if (argc--) {
						out = arg = *argv++;
					}
					else {
						es("Trailing -d\n");
						end_usage();
					}
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'p':
			
				if (str_eq(arg, "printf")) {
					exxx_flag = !arg_bool;
					tick_flag = FALSE;
					usr_flag  = FALSE;
					usr2_flag = FALSE;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'q':

				if (str_eq(arg, "quit")) {
					goto no_more_args;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'r':
			
				if (str_eq(arg, "return")) {
					return_flag = arg_bool;
				}
				else {
					goto unknown_option;
				}
				break;
			
			case 's':
			
				if (str_eq(arg, "sl_fmt")) {
					usr2_flag = arg_bool;
					exxx_flag = !arg_bool;	/* The default. */
					tick_flag = FALSE;
					usr_flag  = FALSE;
				}
				else if (str_eq(arg, "s")) {
					/* Define a path. */
					if (argc--) {
						arg = *argv++;
						#ifdef ENV_MAC
							if(arg[strlen(arg)-1] == ':') {
								name = str_alloc(arg, OBJ_PATH_STRING);
							}
							else {
								name = str_2alloc(arg, ":", OBJ_PATH_STRING);
							}
						#else
							if(arg[strlen(arg)-1] == '\\') {
								name = str_alloc(arg, OBJ_PATH_STRING);
							}
							else {
								name = str_2alloc(arg, "\\", OBJ_PATH_STRING);
							}
						#endif
						path_p = io_new_path_struct(name);
						lst_append(path_p, &arg_search_paths);
					}
					else {
						es("Trailing -s.\n");
						end_usage();
					}
				}
				else {
					goto unknown_option;
				}
				break;
			
			case 't':
			
				if (str_eq(arg, "tick")) {
					tick_flag = arg_bool;
					exxx_flag = !arg_bool;	/* The default. */
					usr_flag  = FALSE;
					usr2_flag = FALSE;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'u':
			
				if (str_eq(arg, "usr")) {
					usr_flag = arg_bool;
					exxx_flag = !arg_bool;	/* The default. */
					tick_flag = FALSE;
					usr2_flag = FALSE;
				}
				else if(str_eq(arg, "-u")) {
					/* Suppress the initial definition of a variable. */
					if (argc--) {
						arg = *argv++;
						/* Put new u_node on global list. */
						mst_undef_arg(arg);
					}
					else {
						es("Trailing -u.\n");
						end_usage();
					}
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'v':
			
				if (str_eq(arg, "v")) {
					err_verbose = TRUE;
				}
				else {
					goto unknown_option;
				}
				break;
				
			case 'x':
				
				if (str_eq(arg, "x")) {
					slc_flag = FALSE;
				}
				else {
					goto unknown_option;
				}
				break;
		
			default:
	
				if (str_eq(arg, "?")) {
					/* Tell the version number and go on. */
					enl(); es(VER_VERSION); enl();
					arg_help();
					break;
				}
	
			unknown_option:
	
				#ifdef PRODUCTION
					if (on_str && arg_prefix(on_str+1, arg)) {
	
						/* Ignore a Sherlock argument. */
						char * s = arg + strlen(on_str) - 1;
						if (*s == '>' && *(s+1) == '>') {
							log_open(s+2);
						}
					}
					else if (off_str && arg_prefix(off_str+1, arg)) {
						/* Ignore a Sherlock argument. */
						;
					}
					else {
						es("ignoring unknown command-line option: ");
						es(arg_bool ? "+" : "-"); es(arg); enl();
					}
				#else
					es("ignoring unknown command-line option: ");
					es(arg_bool ? "+" : "-"); es(arg); enl();
				#endif

			} /* end +- option switch */
		}
		else {
			/* Get a file argument. */
			in = arg;
			break;
		}
	} /* End while. */

	/* Make sure that the output file name was provided. */
	if (in == NULL) {
		es("Missing input file argument.\n");
		end_usage();
	}

	if (out == NULL) {
		es("Missing output file argument.\n");
		end_usage();
	}
	
no_more_args:
	
	/* Remember the input and output paths. */
	if (in != NULL) {
		arg_in_path  = io_new_path_struct(in);
	}
	
	if (out != NULL) {
		arg_out_path = io_new_path_struct(out);
	}
	
	if (argc > 0) {
		arg_save_argv = argv - 1;
		arg_save_argc = argc + 1;
		mst2_init();
		mst_save();
	}
	else {
		mst2_init();
		arg_save_argv = NULL;
	}
	
} /* End arg_do_arg */

#undef debug_arg_do_argv

/*
	Ouput usage message and exit.
*/
static void
arg_help (void)
{
	es("\nusage: spp2 [options] in out\n");
	es("\noptions...\n");
	es("+-d id=value  Define a preprocessor constant\n");
	es("+-exxx        Use exxx routines (default)\n");
	es("+-ftag        Generate FTAG macros\n");
	es("+-f <file>    Use synonym file\n");
	es("+-i           Insert include \"sl.h\" at start of file\n");
	es("+-n           Allow nested comments\n");
	es("+-of          Specify output file\n");
	es("+-printf      Use printf routines\n");
	es("+-return      Allow RETURN_xxx macros\n");
	es("+-sl_fmt      Use sl_fmt routines\n");
	es("+-s <path>    Look for include files in path\n");
	es("+-tick        Insert tick-style macros\n");
	es("+-usr         Use usr output routines\n");
	es("+-u id        Undefine a preprocessor constant\n");
	es("+-v           Verbose output\n");
	es("+-x           Allow single-line comments\n");
	es("+-?           Print the version number and exit\n");		
}

/*
	Resume parsing of arguments.  Return FALSE if none are left.
	
	The previous command-line defines and undefines have 
	already been restored.
*/
bool
arg_init_again(void)
{
	if (arg_save_argv == NULL && arg_paths == NULL) {
		return FALSE;
	}
	else {
		arg_do_argv(0, NULL, "++", "--");
		return TRUE;
	}
}

/*
	Dump various global variables.
*/
void
arg_show (void)
{
	io_path * path;
	
	/*
		Do not put Sherlock macros here: Sherlock has not been initialized!
	*/
	SL_DISABLE();
	
	es("\nGlobal arguments:\n");
	
	es("arg_in_path  -> io_path_name: ");
	es((arg_in_path==NULL) ?  "<NULL>" : arg_in_path -> io_path_name);
	enl();
		
	es("arg_out_path -> io_path_name: ");
	es((arg_out_path==NULL) ? "<NULL>" : arg_out_path -> io_path_name);
	enl();
	
	es("arg_arg_path -> io_path_name: ");
	es((arg_arg_path==NULL) ? "<NULL>" : arg_arg_path -> io_path_name);
	enl();
	
	es("\nGlobal search path arguments:\n");
	
	lst_every(io_path, path, &arg_search_paths) {
		io_show_path(path);
	}
}

/*
	Process a synonym file.
*/
static void
arg_synonyms(void)
{
	char buffer1 [MAX_SYMBOL];
	char buffer2 [MAX_SYMBOL];
	
	SL_DISABLE();

	io_next();
	skip_bl();
	for(;;) {
		if (ch == END_FILE) {
			return;
		}
		else if (isid1(ch)) {
			t_id(&buffer1[0], MAX_SYMBOL);
			skip_bl();
			if (isid1(ch)) {
				t_id(&buffer2[0], MAX_SYMBOL);
				arg_set_syn(&buffer1[0], &buffer2[0]);

				TRACEP("arg_synonyms", es(buffer1); es(" = "); es(buffer2); enl());

				skip_bl();
				if (ch == '\n') {
					io_next();
					skip_bl();
				}
				else {
					error(es("Newline expected"));
					end_abort();
				}
			}
			else {
				error(es("Synonym expected"));
				end_abort();
			}
		}
		else if (ch == ' ' || ch == '\t' || ch == '\n') {
			io_next();
		}
		else if (ch == '#') {
			while (ch != '\n' && ch != END_FILE) {
				io_next();
			}
		}
		else {
			error(es("Identifier or end-of-file expected"));
			end_abort();
		}
	}
}			

/*
	Redefine the names of the Sherlock macros actually generated by SPP.
	Not every Sherlock macro is in the list--only those used by sem.c.
*/
static void
arg_set_syn(reg char * id1, reg char *id2)
{
	SL_DISABLE();
	
	if(str_eq(id1, r_bool)) {
		r_bool = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_char)) {
		r_char = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_double)) {
		r_double = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_float)) {
		r_float = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_int)) {
		r_int = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_long)) {
		r_long = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_ptr)) {
		r_ptr = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_string)) {
		r_string = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_uint)) {
		r_uint = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_ulong)) {
		r_ulong = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, r_void)) {
		r_void = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, s_init)) {
		s_init = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, s_parse)) {
		s_parse = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, statb)) {
		statb = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, statx)) {
		statx = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, tick)) {
		tick = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, tickb)) {
		tickb = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, tickx)) {
		tickx = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, pbtrace)) {
		pbtrace = str_alloc(id2, OBJ_SYN_STRING);
	}
	else if(str_eq(id1, ptrace)) {
		ptrace = str_alloc(id2, OBJ_SYN_STRING);
	}
	else {
		warning(
			es("Synonym entry for "); es(id1);
			es(" has no effect"); enl());
	}
}
