/*
	SPP: Declaration parsing routines.
	
	Source: SPPdcl.c
	Started: September 26, 1987
	Version: March 16, 1996.
*/

#include "SPP.h"

#include "SPPdcl.h"
#include "SPPexp.h"
#include "SPPpar.h"
#include "SPPpr.h"
#include "SPPmain.h"
#include "SPPsem.h"
#include "SPPtok.h"

#include "SPPenv.h"	/* Library additions. */
#include "SPPerr.h"
#include "SPPobj.h"
#include "SPPstr.h"

/*
	Declare routines local to this file.
*/
static void	array		(void);
static void	dcl_enum	(void);
static void	dcl_struct	(void);
static void	fp_list		(en_scope);
static void	function	(en_scope);
static void	initializer	(void);
static void	i_list		(void);
static void	t_list		(void);
static bool	type_head	(en_scope scope);
static en_tokens type_tail	(en_scope scope);
static void	type_ntail	(void);

/* 
	Define flags used in par.c.
	
	These variables are initialized by program()
	so that multiple files can be parsed.
*/
bool break_ok;		/* TRUE if break is allowed.	*/
bool cont_ok;		/* TRUE if break is allowed.	*/
bool in_switch;		/* TRUE if in a switch.			*/

/*
	Declare variables used only by this module.
	
	These variables are initialized by program()
	so that multiple files can be parsed.
*/
static bool	func_def;	/* Inited by program().  TRUE: function DEFINTION was seen. 	*/
static int	prev_dcl;	/* Inited by program().  Number of previous sub-declarations.	*/

/*
	Parse the local variables of a block.
*/
void
block_dcls(void)
{
	SL_NAME(fn_name, "block_dcls");
	TRACETOK(fn_name);

	while (is_type()) {
		dcl(BLOCK_SCOPE);
	}

	STATX(fn_name);
}

/*
	Parse a single declaration or function definition.
	Semicolon ends a declaration but not a function definition.
*/
void
dcl(en_scope scope)
{
	SL_NAME(fn_name, "dcl");
	
	TRACEPB(fn_name, es_scope(scope); ecs(); es_tok(); enl());

	/* Reset semantics for a new declaration. */
	sd_dcl(scope);

	allow_mkeys();
	if (type_head(scope) == TRUE) {
		if (func_def && scope == OUTER_SCOPE) {
			/* We saw a bare function definition. */
			TRACEP(fn_name, es("bare function complete\n"));

			prev_dcl = 0;
			
			STATX(fn_name);
			return;
		}
		else {
			/* We saw a stand-alone struct\union or enum. */
			must(SEMICOLON_TOK);

			STATX(fn_name);
			return;
		}
	}

	/* 2/16/89 */
	for (prev_dcl = 0; ;) {

		type_tail(scope);
		if (scope == OUTER_SCOPE && func_def) {
			/* We ARE at the end of the dcl. */
			if (is(COMMA_TOK)) {
				error(es("Comma following function body ignored"));
				get_token();
			}
			/* 2/16/89 */
			prev_dcl = 0;
			break;
		}
		else if (is(COMMA_TOK)) {
			sd_end(scope, COMMA_TOK);
			get_token();
			/* 2/16/89 */
			prev_dcl++;
		}
		else if (is(SEMICOLON_TOK)) {
			sd_end(scope, SEMICOLON_TOK);
			get_token();
			/* 2/16/89 */
			prev_dcl = 0;
			break;
		}
		else {
			error(
				es("Unexpected end of declaration at ");
				es_tok(); es(", comma or semicolon expected"); enl());

			needend(SEMICOLON_TOK);
			/* 2/16/89 */
			prev_dcl = 0;
			break;
		}
	}

	STATX(fn_name);
}

/*
	A program is nothing but a list of declarations and definitions.
*/
void
program(void)
{
	SL_NAME(fn_name, "program");
	TRACETOK(fn_name);
	
	/*
		Initialize all parsing state flags.
	*/
	prev_dcl		= 0;		/* Used in dcl.c */
	break_ok		= FALSE;	/* Used in par.c */
	cont_ok			= FALSE;	/* Used in par.c */
	in_switch		= FALSE;	/* Used in par.c */
	curly_ok_flag	= FALSE;	/* Used in exp.c */

	/* Get the first token of the program. */
	get_token();

	for (;;) {
		/* Not in function definition yet. */
		func_def = FALSE;

		allow_mkeys();
		if (is_type() || is(ID_TOK)) {
			dcl(OUTER_SCOPE);
		}
		else if (is(EOP_TOK)) {
			STATX(fn_name);
			return;
		}
		else if (is(COMMA_TOK)) {
			error(es("External comma ignored"));
			get_token();
		}
		else if (is(SEMICOLON_TOK)) {
			error(es("External semicolon ignored"));
			get_token();
		}
		else if (is(RCURLY_TOK)) {
			error(es("External '}' ignored"));
			get_token();
		}
		else if (is(LCURLY_TOK)) {
			error(es("Check program structure: { in outer context"));
			get_token();
		}
		else {
			error(es_tok(); es(" ignored: outer declaration expected"); enl());
			get_token();
		}
	}
}

/*
	Skip over any special modifier keywords used by
	Microsoft C or Turbo C.

	Turboc:     asm interrupt
	Microsoft:  fortran 
	Both:       cdecl, far, huge, near, pascal
	Think C:	cdecl, pascal
	
	Do not allow asm when using THINK C!
*/
void
allow_mkeys(void)
{
	char * p;

	TRACETOK("allow_mkeys");

	p = &t_symbol[0];
	while (is(ID_TOK)) {
	
		if (
	
#ifdef ENV_TURBOC
			str_eq(p, "asm") ||
			str_eq(p, "cdecl") ||
			str_eq(p, "far") ||
			str_eq(p, "interrupt") ||
			str_eq(p, "huge") ||
			str_eq(p, "near") ||
			str_eq(p, "pascal")
#endif

#ifdef ENV_MSC
			str_eq(p, "cdecl") ||
			str_eq(p, "far") ||
			str_eq(p, "fortran") ||
			str_eq(p, "huge") ||
			str_eq(p, "near") ||
			str_eq(p, "pascal")
#endif

#ifdef ENV_LSC
			str_eq(p, "cdecl") ||
			str_eq(p, "pascal")
#endif

#ifdef ENV_MPW
			0
#endif

		) {
			get_token();
		}
		else {
			break;
		}
	}
	STATX("allow_mkeys");
}


/*
	Parse a formal paramenter list.
	This may be an old style list (id's only) or a new style list.
	The opening paren has already been read.
*/
static void
fp_list(en_scope scope)
{
	bool no_tail;
	en_scope this_scope;

	SL_NAME(fn_name, "fp_list");
	
	TRACEPB(fn_name, es_scope(scope); ecs(); es_tok(); enl());

	/* Set scope to NULL_SCOPE on recursive entry to this routines. */
	if (	scope == NULL_SCOPE ||
		scope == NEW_FORMAL_SCOPE ||
		scope == OLD_FORMAL_SCOPE) {
		this_scope = NULL_SCOPE;
	}
	else {
		this_scope = NEW_FORMAL_SCOPE;
	}

	while (!is(RPAREN_TOK)) {

		/* Parse the id (old style) or the dcl (new style) */
		allow_mkeys();
		if(is_type() || is(ID_TOK)) {
			sd_dcl(this_scope);
			no_tail = type_head(this_scope);
			if (!no_tail) {
				type_tail(this_scope);
			}
			sd_end(this_scope, SEMICOLON_TOK);
		}
		else if (is(DOTS3)) {
			/* Variable number of parameters. */
			get_token();
		}
		else {
			goto bad_list;
		}

		if (is(COMMA_TOK)) {
			get_token();
		}
		else if (is(RPAREN_TOK)) {
			break;
		}
		else {
			goto bad_list;
		}
	}

	/* Eat the closing paren. */
	get_token();

	STATX(fn_name);
	return;

bad_list:
	error(
		es("Unexpected '"); es_tok();
		es("' seen in formal parameter list"); enl());

	needend(RPAREN_TOK);
	STATX(fn_name);
}

/*
	Parse the old style of declaration of argument types.
*/

static void
t_list(void)
{
	SL_NAME(fn_name, "t_list");
	TRACETOK(fn_name);

	while(is_type()) {
		dcl(OLD_FORMAL_SCOPE);
	}

	STATX(fn_name);
}

/* Parse a list of declarations of array bounds. */
static void
array(void)
{
	SL_NAME(fn_name, "array");
	TRACETOK(fn_name);

	while (is(LBRACK_TOK)) {
		get_token();
		if (is(RBRACK_TOK)) {
			/* [] is valid as an array bounds (sometimes). */
			get_token();
		}
		else {
			con_expr(RBRACK_TOK);
			need(RBRACK_TOK);
		}
	}
	STATX(fn_name);
}

/*
	Parse a function declaration or definition.
	The opening '(' of the formal parameter list has alread been eaten.
*/
void
function(en_scope scope)
{
	SL_NAME(fn_name, "function");
	
	TRACEPB(fn_name, es_scope(scope); ecs(); es_tok(); enl());

	/* Make sure a function definition is valid here. */
	if (scope == STRUCT_SCOPE) {
		error(es("Function defined in struct or union"));
	}
	else if (scope != OUTER_SCOPE) {
		error(es("Nested function definition"));
	}
	else if (prev_dcl > 0) {
		warning(es("Function not expected here"); enl());
	}

	/* Parse the old-style list of parameter types. */
	t_list();

	/*
		We have a function DEFINITION.
		Indicate that '}' ends dcl instead of ';'
		WARNING:  t_list won't work if func_def is TRUE.
	*/
	func_def = TRUE;

	if (!is(LCURLY_TOK)) {
		error(es("Expecting '{' following formal declarations"));
		STATX(fn_name);
		return;
	}

	/* Set the name of the current function. */
	sd_func();

	/* Parse the body of the function. */
	body();

	STATX(fn_name);
}

/*
	Parse an initializer.
*/
static void
initializer(void)
{
	SL_NAME(fn_name, "initializer");
	TRACETOK(fn_name);

	if (is(LCURLY_TOK)) {
		get_token();
		i_list();
		need(RCURLY_TOK);
	}
	else {
		con_expr(SEMICOLON_TOK);
	}

	STATX(fn_name);
}

/*
	Parse a list of initializers.
*/
static void
i_list(void)
{
	SL_NAME(fn_name, "i_list");
	TRACETOK(fn_name);

	for (;;) {

		/* Parse the next item in the list. */
		if (is(LCURLY_TOK)) {
			get_token();
			i_list();
			need(RCURLY_TOK);
		}
		else {
			/* Allow { and } as delimiters. */
			curly_ok_flag  = TRUE;
			con_expr(COMMA_TOK);
			curly_ok_flag = FALSE;
		}

		/* Look for comma and/or '}'. */
		if (is(COMMA_TOK)) {
			get_token();

			/* Comma followed by '}' is valid. */
			if (is(RCURLY_TOK)) {
				STATX(fn_name);
				return;
			}
		}
		else if (is(RCURLY_TOK)) {
			STATX(fn_name);
			return;
		}
		else {
			error(
				es("Unexpected end of initializer at '");
				es_tok(); es("'"); enl());

			get_token();
		}
	}
}

/*
	Parse a type name and the closing paren.
	The opening paren has already been eaten.

	Type names used in casts and sizeof expressions.
*/
void
type_name(void)
{
	SL_NAME(fn_name, "type_name");
	TRACETOK(fn_name);

	type_head(NULL_SCOPE);
	type_ntail();
	need(RPAREN_TOK);

	STATX(fn_name);
}

static void
type_ntail(void)
{
	bool paren_seen = FALSE;

	SL_NAME(fn_name, "type_ntail");
	TRACETOK(fn_name);

	/* Count leading stars. */
	allow_mkeys();
	while (is(STAR_TOK)) {
		get_token();
	}

	/* Issue a warning for identifier. */
	allow_mkeys();
	if (is(ID_TOK)) {
		/* 2/16/89 This is not a user mistake. */
		/* -----
		warning(es("Identifier in type name ignored"); enl());
		----- */
		get_token();
	}
	else if (is(LPAREN_TOK)) {

		get_token();

		/* () or (type) denotes a param list. */
		allow_mkeys();
		if (is(RPAREN_TOK) || is_type()) {
			paren_seen = TRUE;
		}
		else {
			/* The assumed id is in here. */
			type_ntail();
			need(RPAREN_TOK);
		}
	}

	/* The "non-id" part:  the id has already been seen. */
	if (paren_seen || is(LPAREN_TOK)) {
		if (is(LPAREN_TOK)) {
			get_token();
		}
		fp_list(NULL_SCOPE);
	}
	else if (is(LBRACK_TOK)) {
		array();
	}

	STATX(fn_name);
}

/*
	Parse the common type portion of a declaration.

	Return TRUE if no tail is required because of
		1.  struct\union.		struct tag {...};
		2.  enum				enum hue {...}
		3.  old-style formal parameter	f(g);
		4.  3/23/91 a repeated typedef
*/
static bool
type_head(en_scope scope)
{
	char *	old_id;

	SL_NAME(fn_name, "type_head");
	
	TRACEPB(fn_name, es_scope(scope); ecs(); es_tok(); enl());

	/* Compensate for misspelled typedef id's. */
	allow_mkeys();
	if(is(ID_TOK) && !is_type()) {
		/*
			Depending on scope, we might have either
			1.  a "bare" function definition or declaration or
			2.  a misspelled typedef id or
			3.  an old-style formal parameter.
		*/
		old_id = str_alloc(t_symbol, OBJ_TYPDEF_STRING);
		get_token();
		if (	scope == NEW_FORMAL_SCOPE &&
			(is(COMMA_TOK) || is(RPAREN_TOK)) 
		   ) {

			/* An old-style declaration. */
			sd_id(scope, old_id);
			obj_free(old_id);

			STATX(fn_name);
			return TRUE;
		}
		else if (scope == OUTER_SCOPE && is(LPAREN_TOK)) {
			/* Looks like a bare function def or dcl. */
			get_token();
			sd_id(scope, old_id);
			fp_list(scope);
			if (is(LCURLY_TOK) || is_type()) {
				function(scope);
				obj_free(old_id);

				STATX(fn_name);
				return TRUE;
			}
		}
		else if (is(ID_TOK) || is_type()) {
			error(es("Possibly misspelled typedef: "); es(old_id); enl());
		}
		obj_free(old_id);
	}

	while (is_type() && !is(K_STRUCT) && !is(K_UNION) &&!is(K_ENUM)) {
		if (is(K_TYPEDEF)) {
			/* Set typedef flag. */
			sd_typedef(scope);
		}
		else {
			/* Base type or typedef. */
			sd_htype(scope, token);
			if (is(K_ENUM)) {
				break;
			}
		}
		get_token();
	}

	if (is(K_ENUM)) {
		sd_htype(scope, token);
		dcl_enum();
		
		STATX(fn_name);
		return is(SEMICOLON_TOK);
	}
	else if (is(K_STRUCT) || is(K_UNION)) {
		sd_htype(scope, token);
		dcl_struct();

		STATX(fn_name);
		return is(SEMICOLON_TOK);
	}
	else {
		STATX(fn_name);
		return FALSE;
	}
}

/*
	Parse an enum.
*/
static void
dcl_enum(void)
{
	SL_NAME(fn_name, "dcl_enum");
	TRACETOK(fn_name);

	/* Eat the enum. */
	get_token();

	/* get the optional tag. */
	allow_mkeys();
	if (is(ID_TOK)) {
		get_token();
	}

	if (is(LCURLY_TOK)) {

		get_token();
		for (;;) {
			if (is(ID_TOK)) {
				get_token();
			}
			else {
				error(
					es("Identifier expected at '");
					es_tok(); es("' in enum list"); enl());
			}
			if (is(ASSN_TOK)) {
				get_token();
				expr1(RCURLY_TOK);
			}
			if (is(COMMA_TOK)) {
				get_token();
			}
			else {
				break;
			}
		}
		need(RCURLY_TOK);
	}

	STATX(fn_name);
}

/*
	Parse a struct/union definition.
*/
static void
dcl_struct(void)
{
	SL_NAME(fn_name, "dcl_struct");
	TRACETOK(fn_name);

	/* Eat the struct or union. */
	get_token();

	/* get the optional tag. */
	allow_mkeys();
	if (is(ID_TOK)) {
		get_token();
	}

	/* Parse the optional definition. */
	if (is(LCURLY_TOK)) {

		get_token();
		allow_mkeys();
		while (is_type() || is(ID_TOK)) {
			dcl(STRUCT_SCOPE);
		}
		need(RCURLY_TOK);
	}

	/* Parse a possible initializer. */
	if (is(ASSN_TOK)) {
		get_token();
		initializer();
	}

	STATX(fn_name);
}

/*
	Parse the part of the declaration after the type specifier.
	Return the type of the tail or NULL_TOK.
*/
static en_tokens
type_tail(en_scope scope)
{
	en_tokens final_type = NULL_TOK;

	SL_NAME(fn_name, "type_tail");
		
	TRACEPB(fn_name, es_scope(scope); ecs(); es_tok(); enl());

	/* Leading stars. */
	allow_mkeys();
	while (is(STAR_TOK)) {
		sd_type(scope, STAR_TOK);
		get_token();
	}

	/* Optional const qualifier. */
	allow_mkeys();
	if (is(K_CONST)) {
		get_token();
	}

	/* An identifier ALWAYS ends the id part. */
	allow_mkeys();
	if (is(ID_TOK)) {
		sd_id(scope, t_symbol);
		get_token();
	}
	else if (is(LPAREN_TOK)) {
		get_token();

		/* The id should be in here. */
		final_type = type_tail(scope);
		need(RPAREN_TOK);
	}
	else if (scope == STRUCT_SCOPE && is(COLON_TOK)) {

		/* bit field without id. */
		get_token();
		con_expr(SEMICOLON_TOK);
		goto done;
	}
	/* Bug fix: 6/22/89 */
	else if (scope != NEW_FORMAL_SCOPE && scope != NULL_SCOPE) {
		error(
			es("Id expected in declaration. '");
			es_tok(); es("' seen instead"); enl());
		goto done;
	}

	/* The "non-id" part:  the part containing the id has been seen. */
	if (scope == STRUCT_SCOPE && is(COLON_TOK)) {

		/* Bit field after id. */
		get_token();
		con_expr(SEMICOLON_TOK);
		goto done;
	}

	for (;;) {
		if (is(LPAREN_TOK)) {
			get_token();

			/* Parse function declaration or function definition. */
			if (final_type == NULL_TOK) {
				sd_type(scope, LPAREN_TOK);
			}
			fp_list(scope);
			if (is(LCURLY_TOK) || is_type()) {
				function(scope);
				break;
			}
		}
		else if (is(LBRACK_TOK)) {
			/* Parse possible array subscripts. */
			if (final_type == NULL_TOK) {
				sd_type(scope, LBRACK_TOK);
			}
			array();
		}
		else if (is(ASSN_TOK)) {
			/* Parse possible initializer. */
			get_token();
			initializer();
			break;
		}
		
#ifdef ENV_LSC
		/* Allow reference to low-memory globals. */
		else if (is(COLON_TOK)) {
			get_token();
			con_expr(SEMICOLON_TOK);
			goto done;
		}
#endif
		else {
			break;
		}
	}

done:
	/* 3/22/91 */
	STATX(fn_name);
	return final_type;
}
