/*
	SPP: Application-specific environment header.

	source:  SPPenv.h
	started: December 11, 1990.
	version: March 16, 1996.
*/

#ifndef SPPenv_h_
#define SPPenv_h_

#pragma once

/*
	ENV_TYPE completely describes the environemt, one of:
	
	ENV_DOS_TURBO	MS_DOS and Turbo C 2.x compiler.
	ENV_DOS_MSC		MS_DOS and MicroSoft C v6.x compiler.
	ENV_LSC_DIALOG	Macintosh OS, Lightspeed C compiler, argument dialogs.
	ENV_LSC_ARGV	Macintosh OS, Lightspeed C compiler, argv arguments.
	ENV_MPW_TOOL	Macintosh OS, MPW C compiler, MPW tool under MPW.
	ENV_MPW_APPL	Macintosh OS, MPW C compiler, stand-alone application.
	
	The only change that needs to be made to switch environments is to change the
	value of ENV_TYPE.
*/

#define ENV_DOS_TURBOC	1
#define ENV_DOS_MSC		2
#define ENV_LSC_DIALOG	3
#define ENV_LSC_ARGV	4
#define ENV_MPW_TOOL	5
#define ENV_MPW_APPL	6
	
#define ENV_TYPE		ENV_LSC_ARGV

/*
	Define abbreviations determined by ENV_TYPE.
	
	ENV_MACHINE		The machine under which SPP runs, one of:
	
	ENV_DOS			The IBM PC running DOS.
	ENV_TURBOC		The Turbo C 2.x compiler.
	ENV_MSC			MicroSoft C v6.x compiler.
	
	ENV_MAC			The Macintosh OS.
	ENV_LSC			The Lightspeed C compiler.
	ENV_MPW			The MPW C compiler.
	
	The value of ENV_SHERLOCK_DIALOG is also determined by ENV_TYPE.
	
	Do *not change* the following code.  Change the value of ENV_TYPE instead.
*/

#if ENV_TYPE == ENV_DOS_TURBOC
#define ENV_DOS
#define ENV_TURBOC
#undef	ENV_SHERLOCK_DIALOG
#endif

#if ENV_TYPE == ENV_DOS_MSC
#define ENV_DOS
#define ENV_MSC
#undef	ENV_SHERLOCK_DIALOG
#endif

#if ENV_TYPE == ENV_LSC_DIALOG
#define ENV_MAC
#define ENV_LSC
#define ENV_SHERLOCK_DIALOG
#endif

#if ENV_TYPE == ENV_LSC_ARGV
#define ENV_MAC
#define ENV_LSC
#undef ENV_SHERLOCK_DIALOG
#endif

#if ENV_TYPE == ENV_MPW_TOOL
#define ENV_MAC
#define ENV_MPW
#undef	ENV_SHERLOCK_DIALOG
#endif

#if ENV_TYPE == ENV_MPW_APPL
#define ENV_MAC
#define ENV_MPW
#undef ENV_SHERLOCK_DIALOG
#endif

/*
	The following constants are not found in <limits.h>

	ENV_INT_DIGITS 		number of digits in an int, plus 5
	ENV_LONG_DIGITS		number of digits in a long, plus 5
*/
#define ENV_INT_DIGITS	20
#define ENV_LONG_DIGITS	40

/*
	Define the ENV_FAR directive used only by Microsoft C.
*/
#ifdef ENV_MSC
#define ENV_FAR far
#else
#define ENV_FAR
#endif

/*
	Increase stack size for Turbo C.
*/

#ifdef ENV_TURBOC
extern unsigned _stklen = 0xf000;
#endif

#endif /* SPPenv_h_ */
