/*
	SPP: Hold buffer module.
		
	Source:  SPPhold.c.
	Started: February 14, 1991
	Version:
		March 22, 1996.
			Added copy_buf [] and copy_count.
*/

#include "SPP.h"

#include "SPPhold.h"
#include "SPPio.h"
#include "SPPpr.h"

/*
	White space and held output.
*/
char	hws_buf [MAX_HOLD];		/* Held white space. */
int		hws_count;				/* Count of held white space. */
char	hold_buf [MAX_HOLD];	/* Held non-white space. */
int		hold_count = 0;
bool	hold_flag = TRUE;		/* TRUE if holding characters. */
int		hold_ch = FALSE;		/* TRUE if ch should be buffered. */

char	lws_buf [100];			/* Leading white space buffer. */
int		lws_count = 0;			/* Leading white space count. */
int		cnl_count = 0;			/* Consecutive newline count. */

char	copy_buf [MAX_HOLD];	/* Copy of return expression. */
int		copy_count = 0;			/* Number of characters in copy_buf []. */
bool	copy_buf_flag = FALSE;	/* True if copying characters. */

/*
	Print out the hold buffer for debugging.
*/
void
hold_dump(void)
{
	int i;

	/* Output to file. */
	TRACEPN("-file_dump",
		io_puts("\nDump of hold buffer<");
		for (i = 0; i < hold_count; i++) {
			io_puts(pr_ch(hold_buf[i]));
		}
		io_putc('>');
		io_putc('\n'));

	ecnls(2);
	es("Dump of hold buffer:  hold_count = "); eint(hold_count); es("<");
		
	for (i = 0; i < hold_count; i++) {
		es_ch(hold_buf[i]);
	}
	es(">\n\n");
}

/*
	Output ALL held output.
*/
void
hold_flush(void)
{
	int i;

	TRACEP("-hold_flush", hold_dump());

	for (i = 0; hold_count > 0; i++, hold_count--) {
		io_putc(hold_buf [i]);
	}
}

/*
	Re-initialize this module.
*/
void
hold_init_again(void)
{
	
	hws_count	= 0;		/* Held white space.					*/
	hold_count	= 0;		/* Held non-white space					*/
	lws_count	= 0;		/* Leading white space count.			*/
	cnl_count	= 0;		/* Consecutive newline count.			*/
	
	hold_flag	= TRUE;		/* We are holding characters.			*/
	hold_ch		= FALSE;	/* Previous ch should not be buffered.	*/
}

/*
	Discard all held output.
*/
void
hold_kill(void)
{
	TRACEP("-hold_kill", hold_dump());

	hold_count = 0;
}

/*
	Delete all leading newlines and white space from the hold buffer,
	but leave the last leading white space.

	Return a pointer at the first non-blank character or NULL.
*/
char *
hold_nldel(void)
{
	int head, tail, end, i;
	char c;

	TICK("hold_nldel");

	end = hold_count;
	hold_count = 0;
	head = tail = 0;

	/* Scan through the buffer, deleting leading white space. */
	while (head < end) {
		c = hold_buf[head++];
		if (c == '\n') {
			/* Delete any tentatively held white space. */
			tail = 0;
			hold_count = 0;
		}
		else if (c == ' ' || c == '\t') {
			/* Tentatively hold the white space. */
			hold_buf[tail++] = c;
			hold_count++;
		}
		else {
			hold_buf[tail++] = c;
			hold_count++;
			break;
		}
	}

	/* Copy the remainder of the hold buffer. */
	while (head < end) {
		hold_buf[tail++] = hold_buf[head++];
		hold_count++;
	}
	hold_buf[tail] = '\0';

	TRACEPN("-hold_nldel", hold_dump());

	/* Return the first non-white space character of the buffer. */
	for (i = 0; i < hold_count; i++) {
		c = hold_buf[i];
		if (c != ' ' && c != '\t' && c != '\n') {
			return &hold_buf[i];
		}
	}

	return NULL;
}
