/*
	SPP: Initialization and argument processing module.
	
	Source: SPPinit.c
	Started: November 12, 1990
	Version: March 16, 1996.
*/

#include "spp.h"

#include "SPParg.h"
#include "SPPdef.h"
#include "SPPdir.h"
#include "SPPhold.h"
#include "SPPinit.h"
#include "SPPmst.h"
#include "SPPsem.h"
#include "SPPst.h"
#include "SPPtok.h"
#include "SPPver.h"

#include "SPPend.h"	/* Library additions. */
#include "SPPenv.h"
#include "SPPerr.h"
#include "SPPobj.h"
#include "SPPstr.h"

// #include "MacEnv.h"

#include <LIBdialog.h>

void init_args(int * argc, char *** argvp, char * name); // temp
extern char * lib_arg_file_name; // temp

#include <mac_gui.h>	// 3/15/96
#include <LIBes.h>		// 3/15/96

/*
	Define local variables.
*/
static char * init_string = NULL;

/*
	Function prototypes for internal routines.
*/
static void dump_call_back	(void);
static void dump2_call_back	(void);		// 3/15/96
static void	event_call_back	(void);
static void	init_keywords	(void);
static void	keyword			(char * name, int t);
	
/*
	Initialize the environment, Sherlock and all modules.
	
	The argv vector is not used in some environments, so it is possible
	that a compiler could generate a warning to that effect,
	but trying to eliminate that warning would not be wise:
	we definitely want all initialization done by a single routine.
	
	Warning:  if SHERLOCK is defined in *any* file of the object library, make
	*sure* that it is defined here.  Otherwise SPP will crash.
*/
void
init_all(void)
{
	// Initialize the toolbox

		/* Grow the heap to its maximum size. */
		MaxApplZone();	
		
		/* Create more master pointers. */
		MoreMasters();		   		
		MoreMasters();		   		
		MoreMasters();		   		
		MoreMasters();		   
		MoreMasters();
		
		/* Initialize the ROM. */
		InitGraf( (Ptr) &(qd.thePort));
		InitFonts();
		InitWindows();
		InitMenus();
		TEInit();
		InitDialogs(0L);
	
		/* Flush previous events and change the cursor from a watch to an arrow. */
		FlushEvents(everyEvent,0);
		InitCursor();
	
	/* Open the output window. */

	w_mac_init(
		// SL_VERSION_NAME,	/* version name from sl.h.					*/
		FALSE,				/* Initialize the Macintosh ToolBox.		*/
		TRUE,				/* TRUE: output to Sherlock window.			*/
		"SPP LOG",			/* Name of Sherlock window (if used).		*/
		TRUE,				/* TRUE: open Sherlock window.				*/
		TRUE,				/* TRUE: Insert standard menus.				*/
		SHERLOCK_DEFINED,	/* TRUE: Insert Sherlock menu.				*/
		TRUE,				/* TRUE: Insert first item in apple menu.	*/
		"\pAbout SPP...",	/* Title of first item in apple menu.		*/
		&about_dialog,		/* CallBack routine for first apple item.	*/
		TRUE,				/* TRUE: Insert menus immediately.			*/
		&event_call_back,	/* Event callback routine.					*/
		&dump_call_back,	/* User dump callback routine.				*/
		"\pObjects",		/* Name of first user menu item. */
		&dump2_call_back,	/* Second user callback routine. */
		"\pObject Stats",	/* Name of second user menu item. */
		NULL,				/* Third user callback routine. */
		"\p"				/* Name of third user menu item. */
	);

	SL_INIT();

/*
	Since the Sherlock dialog contains a version message,
	print the version only if the Sherlock dialog is not used.
*/

	ecnls(2); es(VER_VERSION); ecnls(2);

	/*
		Initialize the macro symbol table before processing user arguments, so
		that constants can be #define'd or #undef'd from the command line.
		
		Sherlock traces are disabled since Sherlock arguments have not been handled.
	*/
	mst_init();
	st_init();

	/*
		Process user arguments, including Sherlock arguments.
		No Sherlock traces will be in effect before this point.
	*/
	{
		int argc = 0;
		char ** argv = NULL;

		init_args(&argc, &argv, lib_arg_file_name);
		SL_PARSE(argc, argv, "++", "--");
		
		/* Sherlock traces now in effect. */
		arg_do_argv(argc, argv, "++", "--");
		
		TRACEPN("-show_args",
			int i = 0;
			ecnl();
			while (argv[i]) {
				es("argv["); eint(i); es("]: "); es(argv[i]); enl();
				i++;
			}
		);
	}
}

/*
	"Call-back" routines for event processing when Sherlock outputs a line.
*/
static void
event_call_back(void)
{
	if (w_applEvent(0) == FALSE) {	// 3/15/96
		end_quit();
	}
}

static void
dump_call_back(void)
{	
	ecnl();
	obj_show(TRUE, "Sherlock User Report");
	ecnl();
}

static void
dump2_call_back(void)
{
	ecnl();
	obj_stat("Sherlock User Report");
	ecnl();
}

/*
	Re-initialize appropriate globals and return TRUE.
	
	User flags, macro spellings and search paths are not initialized.
*/
bool
init_again(void)
{

	/*
		The following modules are not affected by multiple files:
		
		cvt.c, end.c, env.c, log.c, lst.c, obj.c, pr.c, res.c,
		spp.c, str2.c, sherlock.c, timedate.c.
		
	*/
	
	/*
		Initialize error reporting variables.
		The log stream is not affected by multiple files.
	*/
	err_init_again();
	
	/*
		Initialize the hold buffer.
	*/
	hold_init_again();
	
	/*
		The parser, i.e., dcl.c, par.c and exp.c, is initialized
		by program().  Nothing needs to be done here.
		
		Initialize all other aspects of the compiler itself.
	*/
	
	def_init_again();		/* Macro expansion.		*/
	mst_init_again();
	
	st_end();				/* Symbol table.		*/
	st_init();
	
	dir_init_again();		/* Token processing.	*/
	tok_init_again();
	
	sem_init_again();		/* Semantic routines.	*/
	
	/*
		Reinitialize input and output arguments.
		This should have been done already.

		Search paths remain in effect.
	*/
	arg_in_path		= NULL;
	arg_out_path	= NULL;
	arg_in_strm		= NULL;
	arg_out_strm	= NULL;
	
	/*
		Return TRUE if there are any more files to process.
	*/
	return arg_init_again();
}
