/*
	SPP: Main routine.

	source:  SPPmain.c
	started: October 7, 1985
	Version: March 16, 1996.
*/

#include "SPP.h"

#include "SPParg.h"
#include "SPPdcl.h"
#include "SPPdef.h"
#include "SPPdir.h"
#include "SPPhold.h"
#include "SPPinit.h"
#include "SPPmain.h"
#include "SPPmst.h"
#include "SPPpr.h"
#include "SPPres.h"
#include "SPPsem.h"
#include "SPPnext.h"
#include "SPPst.h"
#include "SPPtok.h"

#include "SPPend.h"	/* Library additions. */
#include "SPPenv.h"
#include "SPPerr.h"
#include "SPPlst.h"
#include "SPPobj.h"
#include "SPPstr.h"
#include "SPPtime.h"

#include <mac_gui.h>	// 3/15/96

/*
	Define local routines.
*/
static int char_val(reg char *s);

void
main(int argc, char **argv)
{
	io_path * path, * path2;
	
	SL_DISABLE();

	/*
		Initialize the machine and Sherlock first.
		It is not possible to provide full error messages during this pass.
	*/
	init_all();
	
	do {
	
		/*
			Open the input and output files.
	
			Open the input file BEFORE opening the input file because
			opening the input file may cause a character to be written.
		*/
		
		if(arg_in_path == NULL) {
			goto done;
		}
		
		/*
			Create an output path for the user if none has been given.
		*/
		if (arg_in_path != NULL && arg_out_path == NULL) {
		
			char out[MAX_FILE_NAME];
		
			strcpy(out, arg_in_path -> io_path_name);
			strcat(out, " .SPP");
			arg_out_path = io_new_path_struct(out);
		}
	
		/* Enter the new input level. */
		io_suspend();
		
		arg_in_strm = io_open_strm(arg_in_path, TRUE);
		if (arg_in_strm == NULL) {
			end_usage();
		}
	
		arg_out_strm = io_create_strm(arg_out_path, TRUE);
		if (arg_out_strm == NULL) {
			end_usage();
		}
		
		err_begin_file(arg_in_path -> io_path_name);
		if (err_verbose) {
			ecnl();
			es("\nCompiling "); es(err_file_name()); es("...\n");
		}
		
		TRACEP("spp_dumps",
			obj_stat("spp_dumps");
			obj_show(TRUE, "spp_dumps"));
	
		/* Insert sl.h if requested. */
		if (insert_flag) {
			io_puts("#include \"sl.h\"\n");
		}
	
		/*
			Start off at a new line and set err_line to zero to compensate.
		*/
		err_line = 0;
		ch = '\n';
		
		TRACEP("spp_arg_show", arg_show());
		
		TRACEP("spp_files",
			es("...\ninput file: ");
			es(arg_in_path -> io_path_name); enl();
			es("output file: "); es(arg_out_path -> io_path_name); enl());
				
		TRACEP("spp_io_show", io_show_all(FALSE));
		
		TRACEP("-spp_tests",
			obj_test();
			mst_test();
		);
		
		/* Parse the program !! */
		program();
		
		/* Empty the hold buffer. */
		hold_flush();
		
		/* Close the input stream. */
		io_close_strm(arg_out_strm);
		
	} while (init_again());
	
	/* Remove the symbol table. */
	st_end();
	
	/* Close all search paths. */
	lst_every(io_path, path, &arg_search_paths) {
		path2 = path;
		io_close_path(&path2);
	}
	
done:

	TRACEP("spp_dumps",
		obj_stat("spp_dumps");
		obj_show(TRUE, "spp_dumps");
		sl_dump());
	
	end_done();
}

/*
	Return the value of a character constant.
*/
static int
char_val(reg char *s)
{
	reg int val;
	
	SL_DISABLE();

	if (*s != '\\') {
		return (int) *s;
	}
	s++;
	switch (*s) {
	case 'b':	return '\b';
	case 'f':	return '\f';
	case 'n':	return '\n';
	case 'r':	return '\r';
	case 't':	return '\t';
	case '\'':	return '\'';
	case '\\':	return '\\';

	default:
		if (*s < '0' || *s > '7') {
			return (int)*s;
		}
		val = 0;
		while (*s >= '0' && *s <= '7') {
			val = val * 8 + (int)*s - '0';
			s++;
		}
		return val;
	}
}

/*
	Return the next token from the input file.
	Set global variables token and t_subtype.

	Note:  Most output is created as a side effect of calling io_next().
*/

/* CAUTION: evaluate value only once! */
#define T_RETURN(value)\
	token = value;\
	TRACEPN(ftag, es("token "); eint(token); ecs(); es_tok(); enl());\
	return;

void
get_token(void)
{
	FTAG("get_token");

	/* Allow sherlock to break here. */
#if ENV_TYPE==ENV_LSC_DIALOG || ENV_TYPE==ENV_LSC_ARGV || ENV_TYPE==ENV_MPW_APPL
	if (w_applEvent(0) == FALSE) {
		end_quit();
	}
#endif
	
	/* Save previous token. */
	/* Bug fix:  11/3/88 */
	if (!con_flag) {
		last_tok = token;
	}

	/* Reset first_token_in_line flag. */
	first_tok = FALSE;

	/* Output the buffer. */
	hold_flush();

	/* Get the next token. */
	get_t1(TRUE);

	TRACEP("get_token",
		es("file: "); es(err_file_name());
		es(", line: "); elong(err_line);
		es(", tok: "); es_tok();
		es(" = "); eint(token); enl());
}

/*
	Same as get_token but do not allow expansion.
*/
void
get_xtoken(void)
{
	FTAG("get_xtoken");
	TICK(ftag);

	/* Save previous token. */
	last_tok = token;

	/* Reset first_token_in_line flag. */
	first_tok = FALSE;

	/* Output the buffer. */
	hold_flush();

	/* Get the next token. */
	get_t1(FALSE);

	TRACEP(ftag, eret();
		es_tok(); es(" = "); eint(token); ecs();
		es("last_tok: "); eint(last_tok); ecs();
		es("ch: "); es_ch(ch); enl());
}

/*
	Kludgy buffer for held id's after a string.
*/
static bool id_flag = FALSE;
static char id_buf [MAX_SYMBOL];

void
get_t1(bool expand_flag)
{
	FTAG("get_t1");

	int i;
	bool old_mflag;
	
	TRACEP(ftag, es("line "); elong(err_line); enl());

	*t_symbol = '\0';
	t_subtype = 0;

	/* Kludge required by string catenation. */
	if (id_flag) {
		id_flag = FALSE;
		strcpy(&t_symbol[0], &id_buf[0]);
		T_RETURN(ID_TOK);
	}

	/*
		We branch to the 'rescan' label whenever a construct is seen
		that does not result directly in a token being returned, i.e.,
		for macros, PP directives and whitespace.
	*/

rescan:

	TRACEP("get_t1_at_rescan", es("ch: "); es_ch(ch); enl());

	/* Special case for "wide" characters and strings. */
	if (ch == 'L') {
		io_next();
		if (isid2(ch)) {
			io_pushback(ch);
			ch = 'L';
		}
	}

	if (isid1(ch)) {
		old_mflag = m_flag;
		t_id(&t_symbol[0], MAX_SYMBOL);
		token = is_reserved(t_symbol, t_length);
		if (is(NULL_TOK)) {
			token = ID_TOK;
			if (	expand_flag &&
				outer_expand(&t_symbol[0], old_mflag)
			   ) {
				goto rescan;
			}
			else {
				T_RETURN(token);
			}
		}
		else {
			T_RETURN(token);
		}
	}

	switch (ch) {

	TICK("get_token1");

	case ' ':
	case '\t':
		/* Accumulate non-leading white space. */
		hws_count = 0;
		while (ch == ' ' || ch == '\t') {
			hws_buf[hws_count++] = ch;
			io_next();
		}
		goto rescan;

	case '\n':

		/* A kludge: don't skip past newlines in eval(). */
		if (con_flag) {
			con_flag = FALSE;
			TRACEPN("get_token", es("con_flag: NULL_TOK\n"));
			T_RETURN(NULL_TOK);
		}

        /* Allow user to abort here. */
		
		io_next();
		bump_line();

		/*	Accumulate leading white space.
			Do not change the count unless there is some.
			Set the consecutive newline count.
		*/
		cnl_count = 1;
		while (ch == ' ' || ch == '\t' || ch == '\n') {
			if (ch == '\n') {
				cnl_count++;
				bump_line();
				io_next();
			}
			else if (!con_flag && (ch == ' ' || ch == '\t')) {
				lws_count = 0;
				while (ch == ' ' || ch == '\t') {
					lws_buf [lws_count++] = ch;
					io_next();
				}
			}
		}
		begin_line(TRUE);
		goto rescan;

	case '#':
		error(es("Unexpected # ignored"));
		io_next();
		goto rescan;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':

		T_RETURN(t_number(FALSE));

	case '.':
		io_next();
		if (ch >= '0' && ch <= '9') {
			T_RETURN(t_number(TRUE));
		}
		else if (ch == '.') {
			io_next();
			if (ch == '.') {
				io_next();
				T_RETURN(DOTS3);
			}
		}
		else {
			T_RETURN(DOT_TOK);
		}

	/* -----
	case '"':
		t_string(t_symbol, MAX_SYMBOL, FALSE);
		T_RETURN(STRING_TOK);
	----- */

	case '"':

		/*	Do concatenation of string literals here.
			Note that a macro may expand into several strings.
		*/
		t_symbol[0] = '"';
		for(i = 1;;) {
			if (ch == ' ' || ch == '\t') {
				io_next();
			}
			else if (ch == '"') {
				t_string(&t_symbol[i], MAX_SYMBOL-i, FALSE);
				i = strlen(&t_symbol[0]);
			}
			else if (isid1(ch)) {
				old_mflag = m_flag;
				t_id(&id_buf[0], MAX_SYMBOL);
				if (!outer_expand(&id_buf[0], old_mflag)) {
					/*
						Not a macro.
						Restore the id and return the
						stirng.
					*/
					id_flag = TRUE;
					break;
				}
			}
			else {
				break;
			}
		}
		t_symbol[i++] = '"';
		t_symbol[i]   = '\0';
		T_RETURN(STRING_TOK);

	case '\'':
		t_string(t_symbol, MAX_SYMBOL, FALSE);
		t_value = (long) char_val(t_symbol);
		T_RETURN(CHAR_TOK);

	/*
		We must be VERY careful about exactly when we switch from one
		input file to the next.  This is the place. 
	*/

	case END_FILE:
	
		TRACEP("end_file",
			es("io_in_strm_level: "); eint(io_in_strm_level); enl());

		if (con_flag) {
			con_flag = FALSE;
			T_RETURN(NULL_TOK);
		}

		/* Switch input streams. */
		io_close_strm(io_strm_in_list);
		io_resume();
		
		if (io_in_strm_level == -1) {
			T_RETURN(EOP_TOK);
		}
		else {
			goto rescan;
		}

	case '=':	/* = or == */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(EQUAL_TOK);
		}
		else {
			T_RETURN(ASSN_TOK);
		}


	case '+':	/* + or ++ or += */

		io_next();
		if (ch == '+') {
			io_next();
			T_RETURN(INC_TOK);
		}
		else if (ch == '=') {
			io_next();
			T_RETURN(PLUS_ASSN_TOK);
		}
		else {
			T_RETURN(PLUS_TOK);
		}


	case '-':	/* - or -- or -> */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(MINUS_ASSN_TOK);
		}
		else if (ch == '-') {
			io_next();
			T_RETURN(DEC_TOK);
		}
		else if (ch == '>') {
			io_next();
			T_RETURN(ARROW_TOK);
		}
		else {
			T_RETURN(MINUS_TOK);
		}


	case '*':	/* * or *= */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(STAR_ASSN_TOK);
		}
		else {
			T_RETURN(STAR_TOK);
		}


	case '/':	/* comment or / or /= */

		io_next();
		if (ch == '*') {
			io_next();
			skip_comment();
			goto rescan;
		}
		else if (slc_flag && ch == '/') {
			/* 8/3/89 MicroSoft single-line comment. */
			skip_past();
			goto rescan;
		}
		else if (ch == '=') {
			io_next();
			T_RETURN(DIV_ASSN_TOK);
		}
		else {
			T_RETURN(DIV_TOK);
		}


	case '%':	/* % or %= */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(MOD_ASSN_TOK);
		}
		else {
			T_RETURN(MOD_TOK);
		}


	case '>':	/* > or >= or >> or >>= */

		io_next();
		if (ch == '>') {
			io_next();
			if (ch == '=') {
				io_next();
				T_RETURN(RSHIFT_ASSN_TOK);
			}
			else {
				T_RETURN(RSHIFT_TOK);
			}
		}
		else if (ch == '=') {
			io_next();
			T_RETURN(GE_TOK);
		}
		else {
			T_RETURN(GT_TOK);
		}


	case '<':	/* < or or <= or << or <<= */

		io_next();
		if (ch == '<') {
			io_next();
			if (ch == '=') {
				io_next();
				T_RETURN(LSHIFT_ASSN_TOK);
			}
			else {
				T_RETURN(LSHIFT_TOK);
			}
		}
		else if (ch == '=') {
			io_next();
			T_RETURN(LE_TOK);
		}
		else {
			T_RETURN(LT_TOK);
		}


	case '!':	/* ! or != */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(NE_TOK);
		}
		else {
			T_RETURN(NOT_TOK);
		}
			

	case '|':	/* | or |= or || */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(OR_ASSN_TOK);
		}
		else if (ch == '|') {
			io_next();
			T_RETURN(LOR_TOK);
		}
		else {
			T_RETURN(OR_TOK);
		}


	case '&':	/* & or &= or && */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(AND_ASSN_TOK);
		}
		else if (ch == '&') {
			io_next();
			T_RETURN(LAND_TOK);
		}
		else {
			T_RETURN(AND_TOK);
		}


	case '^':	/* ^ or ^= */

		io_next();
		if (ch == '=') {
			io_next();
			T_RETURN(XOR_ASSN_TOK);
		}
		else {
			T_RETURN(XOR_TOK);
		}


	case '?':	io_next();	T_RETURN(QUESTION_TOK);
	case ':':	io_next();	T_RETURN(COLON_TOK);

	case '~':	io_next();	T_RETURN(TILDE_TOK);
	case ',':	io_next();	T_RETURN(COMMA_TOK);
	case '(':	io_next();	T_RETURN(LPAREN_TOK);
	case ')':	io_next();	T_RETURN(RPAREN_TOK);
	case '[':	io_next();	T_RETURN(LBRACK_TOK);
	case ']':	io_next();	T_RETURN(RBRACK_TOK);
	case '{':	io_next();	T_RETURN(LCURLY_TOK);
	case '}':	io_next();	T_RETURN(RCURLY_TOK);
	case ';':	io_next();	T_RETURN(SEMICOLON_TOK);

	default:

		error(es("Invalid character ignored: "); echar(ch); enl());

		io_next();
		goto rescan;
	}
}

#undef T_RETURN

/*
	Do beginning of line processing.
	Look for preprocessor directives only if flag is TRUE.
*/
void
begin_line(bool flag)
{
	TRACEP("begin_line", es(err_file_name()); ecs(); elong(err_line); enl());

	skip_ws(TRUE);

	/* PP directives are not allowed as the result of macro expansion. */
	if (!con_flag && flag && ch == '#' && !m_flag) {
		io_next();
		do_pp();
	}

	first_tok = TRUE;
}

/*
	Do end of line processing.
*/
void
bump_line(void)
{
	SL_DISABLE();
	
	err_line++;
	
	TRACE("bump_line",
		if ((err_line % 50) == 0) {
			help(es("bump_line"); enl());
		});
}
