/*
	SPP: routines that get the next input character.
	
	Source:  SPPnext.c
	Started: May 3, 1991.
	Version: March 16, 1996.
*/

#include "SPP.h"

#include "SPPdef.h"
#include "SPPhold.h"
#include "SPPnext.h"
#include "SPPpr.h"
#include "SPPtok.h"

#include "SPPenv.h"	/* Library additions. */
#include "SPPerr.h"
#include "SPPio.h"
#include "SPPobj.h"
#include "SPPstr.h"

/*
	Function prototypes of local routines.
*/
static void	io_n1 (void);

/*
	Push c back into the file input stream, not the macro input stream.
*/
static int file_put_back = -1;

void
io_fpb(int c)
{
	file_put_back = c;
}

/*
	Set the global variable ch to the next character from the input stream.
	This is the most often called routine in the whole program.

	This routine eliminates all  '\r' and backslash-newlines.
	Thus, '\r' and backslash-newline may appear ANYWHERE in the program.
	
	This function also copies all input characters either to the hold
	buffer or directly to the output.  Only actual source characters
	are copied:  characters from included files or macro expansion
	characters are not copied.
	
	This routine is complicated by the complexities of macro expansion,
	which are handled in one pass so as to make the copying of source
	characters bullet proof.
*/

static int io_push_back = -1;

void
io_next(void)
{
	FTAGV("-io_next_v");
	SL_DISABLE();
	
	for(;;) {
			
		/* sysn1() filters out all '\r' characters. */
		io_n1();

		/* Intercept backslash newline combinations here. */
		if (ch == '\\') {
			io_n1();
			if (ch == '\n') {
				err_line++;
			}
			else {
				io_push_back = ch;
				ch = '\\';
				
				// TRACE(ftagv, echar(ch));
				goto done;
			}
		}
		else {
			// TRACE(ftagv, echar(ch));
			goto done;
		}
	}
	
done:

	/* 3/22/96 */
	if (copy_buf_flag) {
		/* Copy characters to the copy buffer. */
		if (copy_count < MAX_HOLD) {
			copy_buf [copy_count++] = ch;
		}
		else if (copy_count == MAX_HOLD) {
			warning(es("return expression truncated!"); enl());
			copy_count++;
		}
	}
}

static void
io_n1(void)
{
	FTAG("-io_n1");
	FTAGV("-io_n1_v");
	
	reg int			n;
	reg io_strm *	stream_p;

	/* Priority 1: A pushed back character. */
	if (io_push_back != -1) {
		ch = io_push_back;
		// TRACEP(ftag, es("pushback: "); es_ch(ch); enl());
		io_push_back = -1;
		goto done;
	}

	/* Buffer or output the *previous* character. */
	if (!m_flag && !hold_flag && hold_ch) {
		// TRACEPN(ftag, es("immediate out "); es_ch(ch); enl());
		io_putc(ch);
	}
	else if (!m_flag && hold_ch) {
		if (hold_count >= MAX_HOLD) {
			warning(es("Hold buffer overflow"); enl());
			hold_flush();
		}
		hold_buf [hold_count++] = ch;
		// TRACEPN(ftag, es("buffers "); es_ch(ch); enl());
	}

loop:

	/* Priority 2:  The rescan buffer. */
	if (m_flag) {
		ch = *p_rescan++;
		if (ch) {
			// TRACEP(ftag, es("rescan ch: "); es_ch(ch); enl());
			goto done;
		}
		m_flag = FALSE;

		/*
			Priority 2A:  The trailing character.
			This is NOT part of the rescan buffer.
		*/
		if (file_put_back != -1) {
			ch = file_put_back;
			// TRACEP(ftag, es("fpb ch: "); es_ch(ch); enl());
			file_put_back = -1;
			goto done;
		}
		/* FALL THROUGH. */
	}

	/* Priority 3:  The current file. */
	
	stream_p = io_strm_in_list;
	ASSERT(stream_p);
	
	if (stream_p -> io_strm_bufc > 0) {
		stream_p -> io_strm_bufc--;
		ch = *(stream_p -> io_strm_bufp)++;
		#if 0
		TRACEP(ftagv,
			es("ch: "); es_ch(ch);
			es(", bufp: "); eptr(stream_p -> io_strm_bufp);
			es(", bufc: "); elong(stream_p -> io_strm_bufc);
			es(", file: ");
			es(stream_p -> io_strm_file -> io_file_path -> io_path_name));
		#endif
		goto done;
	}

	/* A file or local buffer is empty. */
	
	switch(stream_p -> io_strm_file -> io_file_stat) {

	case IO_EOF:
		/* Continue returning END_FILE until the file is closed. */
		TICK("io_next_eof2");
		ch = END_FILE;
		return;

	case IO_INPUT:
	
		n = io_read(
			stream_p -> io_strm_file,
			stream_p -> io_strm_buffer,
			IO_BUF_SIZE);
			
		if (n > 0) {
			stream_p -> io_strm_bufc = n - 1;
			stream_p -> io_strm_bufp = stream_p -> io_strm_buffer;
			ch = *(stream_p -> io_strm_bufp)++;
			goto done;
		}
		else {
			TICK("io_next_eof1");
			stream_p -> io_strm_file -> io_file_stat = IO_EOF;
			ch = END_FILE;
			return;
		}

	default:
		
		es("\nio_strm_in_list = "); eptr(io_strm_in_list); enl();
		fatal(es("io_n1: Bad file status field"));
	}

done:
	
#ifdef ENV_MAC

	/* The mac has a weird sense of humor. */
	if (ch == '\r') {
		ch = '\n';
	}
	
#else

	/* Never pass '\r' to the rest of the program. */
	if (ch == '\r') {
		goto loop;
	}

#endif

	/*
		Only buffer non-macro characters in the current file.
		Do not buffer any character if hold_flag is TRUE.
	*/
	hold_ch = !m_flag && io_in_strm_level == 0;

	TRACEP(ftag, eret(); es_ch(ch); enl());
}

/*
	Push back c so that io_next() will return it next.
*/
void
io_pushback( int c)
{
	io_push_back = c;
}

/*
	Resume the previous level of input AFTER a stream has been closed.
	Restore the global variable ch.
*/
void
io_resume(void)
{
	FTAG("io_resume");

	io_strm * stream;
	
	SL_DISABLE();

	/* Pop back one level. */
	io_in_strm_level--;
	stream = io_strm_in_list;
	
	if (io_in_strm_level == -1 && stream != NULL) {
		fatal(es("io_resume: io_in_strm_level, input list mismatch 1"));
	}
	
	if (io_in_strm_level != -1 && stream == NULL) {
		fatal(es("io_resume: io_in_strm_level, input list mismatch 2"));
	}

	if (io_in_strm_level == -1) {
		ch = END_FILE;
		TRACEPN(ftag, es("io_in_strm_level: -1, ch=END_FILE\n"));
	}
	else {
	
		/* Restore the stream's name, line number and ch. */
		err_begin_file(stream->io_strm_file->io_file_path->io_path_name);
		err_line = stream -> io_strm_file -> io_file_line;
		ch = (unsigned char) stream -> io_strm_file -> io_file_lastc;
		
		TRACEPN(ftag,
			es("io_in_strm_level: "); eint(io_in_strm_level);
			es(", file: "); es(stream -> io_strm_file -> io_file_path -> io_path_name);
			es(", line: "); elong(err_line);
			es(", bufc: "); elong(stream -> io_strm_bufc);
			es(", ch: "); es_ch(ch); enl());
	}
}

/*
	Suspend the previous input stream BEFORE opening another stream.
	Force a newline on the new stream and set err_line to 0.
*/
void
io_suspend(void)
{
	FTAG("io_suspend");

	io_strm * stream;
	
	stream = io_strm_in_list;

	TRACEP(ftag,
		es("old io_in_strm_level: "); eint(io_in_strm_level); enl());

	if (m_flag == TRUE) {
		m_flag = FALSE;
		error(es("Macro expansion truncated following #include"));
	}

	/* Save information about the current level on the stack. */
	if (io_in_strm_level != -1) {
		if (stream == NULL) {
			fatal(es("io_suspend: Can't happen"));
		}
		stream -> io_strm_file -> io_file_line = err_line;
		stream -> io_strm_file -> io_file_lastc = (int) ch;
	}

	/* Enter a new level. */
	io_in_strm_level++;

	/*
		Force a '\n' as the first character and set err_line to 0.
		This will make preprocessor directives  valid on the first line.

		Since io_next() is not called to get this character,
		it won't be buffered or output.
	*/
	ch = '\n';
	err_line = 0;
}
