/*
	SPP: printing routines for debugging.
	
	source:  SPPpr.c
	started: May 19, 1986
	version: March 16, 1996.
*/

#include "SPP.h"

#include "SPPpr.h"
#include "SPPtok.h"

#include "SPPenv.h"	/* Library additions. */
#include "SPPerr.h"
#include "SPPio.h"
#include "SPPstr.h"

// Private prototypes

static char * pr_scope		(en_scope scope);
static char * pr_short_fn	(char * file_name);
static char * pr_tok		(void);

/*
	WARNING!!  These two tables must change if enum.h changes.
*/
static char * scope_tab [] = {
	"NULL_SCOPE", "OUTER_SCOPE", "OLD_FORMAL_SCOPE",
	"NEW_FORMAL_SCOPE", "STRUCT_SCOPE", "BLOCK_SCOPE"
};

static char * ENV_FAR tok_tab [] = {
	"<NULL>", "<ERROR>",

	"auto", "char", "const", "double", "enum", "extern",
	"float", "int", "long", "noalias", "register", "short",
	"static", "typedef", "signed", "struct", "union",
	"unsigned", "void", "volatile",

	"break", "case",
	"continue", "default", "do", "else", "for",
	"goto",	"if", "return", "switch", "while",

	"entry", "sizeof",

	",", "<newline>", ";", "[", "{", "(", "]", "}", ")",

	":", "?",

	"ARRAY", "->", ".", "&&", "||", ",",

	"=",
	"&=", "/=", "<<=", "-=", "%=",
	"|=", "+=", ">>=", "*=", "^=",

	"&", "|", "+", "*", "^",
	"/", "<<", "-", "%", ">>",

	"==", ">=", ">", "<=", "<", "!=",

	"--", "++", "!", "~",

	"(cast)", "post--", "post++", "pre--", "pre++",
	"&", "-", "+", "*", "call",
	
	"character", "end-of-file",
	"identifier", "integer", "float", "long integer",
	"string",
	"unsigned long", "long double",
	"...", "label", "bool"
};

void
es_ch(int c)
{
	es(pr_ch(c));
}

void
es_op (en_tokens tok)
{
	es(pr_op(tok));
}

void
es_scope (en_scope scope)
{
	es(pr_scope(scope));
}

void
es_short_fn (char * file_name)
{
	es(pr_short_fn(file_name));
}

void
es_str (char *s)
{
	while (*s) {
		pr_ch(*s);
		++s;
	}
}

void
es_tok (void)
{
	es(pr_tok());
}

/*
	Return the print string of a character.
*/
static char pdefault[3] = "@";

char *
pr_ch(int c)
{
	SL_DISABLE();
	
	switch (c) {
	case ARG_FLAG:		return "ARG:";
	case POUND_FLAG:	return "POUND:";
	case CONCAT_FLAG:	return "<CONCAT_FLAG>";
	case EORT:			return "<EORT>";
	case EXPAND_OFF:	return "<EXPAND_OFF>";
	case '\0':			return "<\\0>";
	case '\t':			return "<TAB>";
	case '\n':			return "<NL>";
	case '\r':			return "<CR>";
	case '\a':			return "<ALERT>";
	case END_FILE:		return "<END_FILE>";
	default:
		pdefault[0] = c;
		pdefault[1] = '\0';
		return pdefault;
	}
}

/*
	Return the print string of tok.
	Do not assume that t_symbol[] refers to tok.
*/
char *
pr_op(en_tokens tok)
{
	SL_DISABLE();
	
	if ((int)tok >= 0 && (int)tok <= LAST_ENUM_TOK) {
		return tok_tab [(int)tok];
	}
	else {
		return "BAD TOKEN";
	}
}

/*
	Given a static string, return the short file name of the string.
*/
char *
pr_short_fn(char * file_name)
{
	reg int i, n;
	
#ifdef ENV_MAC
		
	/* Scan backwards looking for a colon. */
	n = strlen(file_name)-2;
	for (i = n; i >= 0; i--) {
		if (file_name[i] == ':') {
			return file_name + i + 1;
		}
	}
	return file_name;
	
#else
	return file_name;
#endif

}

char *
pr_scope(en_scope scope)
{
	SL_DISABLE();
	
	return scope_tab [scope];
}

/*
	Return the print string of the current token.
	Assume the t_symbol refers to the global token.
	
	Warning:  do not use strncpy here!
*/
static char buffer [200];
static char buffer2 [ENV_LONG_DIGITS];

char *
pr_tok(void)
{
	SL_DISABLE();
	
	if (is(ID_TOK)) {
		if (t_length >= 200-5) {
			fatal(es("truncated identifier: "); es(t_symbol); enl());
			return "";
		}
		else {
			strcpy(buffer, t_symbol);
			return buffer;
		}
	}
	else if (is(STRING_TOK)) {
		if (t_length >= 200-5) {
			fatal(es("truncated string: "); es(t_symbol); enl());
			return "";
		}
		else {
			strcpy(buffer, t_symbol);
			return buffer;
		}
	}
	else if (is(INT_TOK)) {
		char buffer [100];
		return cvt_long(buffer, 100, t_value);
	}
	else if (token < LABEL_TOK) {
		return  tok_tab [(int)token];
	}
	else {
		return "BAD TOKEN";
	}
}
