/*
	SPP: reserved word handling.
	
	Source:  SPPres.c
	Started: September 28, 1987
	Version: March 16, 1996.
*/

#include "SPP.h"

#include "SPPres.h"
#include "SPPtok.h"

#include "SPPenv.h"
#include "SPPstr.h"

/*
	Spellings of ALL Sherlock macros.
*/
static char * mn_tab [] = {

	"FTAG",

	"RETURN_BOOL",	"RETURN_CHAR",	 "RETURN_DOUBLE",
	"RETURN_FLOAT",	"RETURN_INT",	 "RETURN_LONG",
	"RETURN_PTR",	"RETURN_STRING", "RETURN_UINT",
	"RETURN_ULONG",	"RETURN_VOID",

	"SL_CLEAR",	"SL_DISABLE",	"SL_DUMP",	"SL_INIT",
	"SL_NAME",	"SL_OFF",		"SL_ON",	"SL_PARSE",

	"STAT",		"STATB",	"STATX",
	"TICK",		"TICKB",	"TICKN",	"TICKX",
	"TRACE",	"TRACEB",	"TRACEN",	"TRACEX",
	"TRACEP",	"TRACEPB",	"TRACEPN",	"TRACEPX",

	NULL
};

/*
	Spellings of macros generated by the code generators.
*/
char *	r_bool	= "RETURN_BOOL";
char *	r_char	= "RETURN_CHAR";
char *	r_double= "RETURN_DOUBLE";
char *	r_float	= "RETURN_FLOAT";
char *	r_int	= "RETURN_INT";
char *	r_long	= "RETURN_LONG";
char *	r_ptr	= "RETURN_PTR";
char *	r_string= "RETURN_STRING";
char *	r_uint	= "RETURN_UINT";
char *	r_ulong	= "RETURN_ULONG";
char *	r_void	= "RETURN_VOID";

#ifdef ENV_DOS
char *	s_init	= "SL_INIT";
#else
char *	s_init	= "SL_MAC_INIT";
#endif

char *	s_parse	= "SL_PARSE";

/*
	Only the macros used in sem.c are defined here.
*/

char *	statb	= "STATB";
char *	statx	= "STATX";
char *	tick	= "TICK";
char *	tickb	= "TICKB";
char *	tickx	= "TICKX";

char *	pbtrace	= "TRACEPB";
char *	pxtrace	= "TRACEPX";
char *	ptrace	= "TRACEP";

/*
	Define the reserved word tables using macros.
*/
typedef struct entry_struct entry_node;

static struct entry_struct {
	char *	e_string;
	int		e_arg1;		/* en_token value. */
};

typedef struct tab_struct tab_node;

static struct tab_struct {
	entry_node *	t_entry;
	short 			t_strlen;
	short 			t_entries;
};

#define table(name, n) tab_node ENV_FAR name [n+1] = {
#define sub_tab(name, strlen, n_entries) {name, strlen, n_entries}
#define end_tab {0L, 0}}

#define header(label,n_entries) entry_node ENV_FAR label [n_entries] = {
#define key(string,val)	{string,val}
#define end_h 		}

/*
	Reserved word tables (One entry per reserved word.)

	Each table is made up of zero or more subtables,
	one subtable for each length.
	Subtables are arranged in increasing length.
	Within each subtable, entries are arranged alphabetically.
*/

/* ----- A TABLE ----- */
#define a4_n 1
	header(a4_tab, a4_n)
	key("auto", K_AUTO)
	end_h;

	table(a_tab, 1)
	sub_tab(a4_tab, 4, a4_n),
	end_tab;

/* ----- B TABLE ----- */
#define b5_n 1
	header(b5_tab, b5_n)
	key("break", K_BREAK)
	end_h;

	table(b_tab, 1)
	sub_tab(b5_tab, 5, b5_n),
	end_tab;

/* ----- C TABLE ------ */
#define c4_n 2
	header(c4_tab, c4_n)
	key("case", K_CASE),
	key("char", K_CHAR),
	end_h;

#define c5_n 1
	header(c5_tab, c5_n)
	key("const", K_CONST)
	end_h;

#define c8_n 1
	header(c8_tab, c8_n)
	key("continue", K_CONTINUE)
	end_h;

	table(c_tab, 3)
	sub_tab(c4_tab, 4, c4_n),
	sub_tab(c5_tab, 5, c5_n),
	sub_tab(c8_tab, 8, c8_n),
	end_tab;

/* ----- D TABLE ----- */
#define d2_n 1
	header(d2_tab, d2_n)
	key("do", K_DO),
	end_h;

#define d6_n 1
	header(d6_tab, d6_n)
	key("double", K_DOUBLE)
	end_h;

#define d7_n 1
	header(d7_tab, d7_n)
	key("default", K_DEFAULT)
	end_h;

	table(d_tab, 3)
	sub_tab(d2_tab, 2, d2_n),
	sub_tab(d6_tab, 6, d6_n),
	sub_tab(d7_tab, 7, d7_n),
	end_tab;

/* ----- E TABLE ----- */
#define e4_n 2
	header(e4_tab, e4_n)
	key("else", K_ELSE),
	key("enum", K_ENUM),
	end_h;

#define e5_n 1
	header(e5_tab, e5_n)
	key("entry", K_ENTRY)
	end_h;

#define e6_n 1
	header(e6_tab, e6_n)
	key("extern", K_EXTERN)
	end_h;

	table(e_tab, 3)
	sub_tab(e4_tab, 4, e4_n),
	sub_tab(e5_tab, 5, e5_n),
	sub_tab(e6_tab, 6, e6_n),
	end_tab;

/* ----- F TABLE ----- */
#define f3_n 1
	header(f3_tab, f3_n)
	key("for", K_FOR)
	end_h;

#define f5_n 1
	header(f5_tab, f5_n)
	key("float", K_FLOAT)
	end_h;

	table(f_tab, 2)
	sub_tab(f3_tab, 3, f3_n),
	sub_tab(f5_tab, 5, f5_n),
	end_tab;

/* ----- G TABLE ----- */
#define g4_n 1
	header(g4_tab, g4_n)
	key("goto", K_GOTO)
	end_h;

	table(g_tab, 1)
	sub_tab(g4_tab, 4, g4_n),
	end_tab;

/* ----- I TABLE ----- */
#define i2_n 1
	header(i2_tab, i2_n)
	key("if", K_IF)
	end_h;

#define i3_n 1
	header(i3_tab, i3_n)
	key("int", K_INT)
	end_h;

	table(i_tab, 2)
	sub_tab(i2_tab, 2, i2_n),
	sub_tab(i3_tab, 3, i3_n),
	end_tab;

/* ----- L TABLE ----- */
#define l4_n 1
	header(l4_tab, l4_n)
	key("long", K_LONG)
	end_h;

	table(l_tab, 1)
	sub_tab(l4_tab, 4, l4_n),
	end_tab;

/* ----- N TABLE ----- */
#if 0 // noalias was removed from the draft standard
	#define n7_n 1
		header(n7_tab, n7_n)
		key("noalias", K_NOALIAS)
		end_h;
	
		table(n_tab, 1)
		sub_tab(n7_tab, 7, n7_n),
		end_tab;
#endif

/* ----- R TABLE ----- */
#define r6_n 1
	header(r6_tab, r6_n)
	key("return", K_RETURN)
	end_h;

#define r8_n 1
	header(r8_tab, r8_n)
	key("register", K_REGISTER)
	end_h;

	table(r_tab, 2)
	sub_tab(r6_tab, 6, r6_n),
	sub_tab(r8_tab, 8, r8_n),
	end_tab;

/* ----- S TABLE ----- */
#define s5_n 1
	header(s5_tab, s5_n)
	key("short", K_SHORT),
	end_h;

#define s6_n 5
	header(s6_tab, s6_n)
	key("signed", K_SIGNED),
	key("sizeof", K_SIZEOF),
	key("static", K_STATIC),
	key("struct", K_STRUCT),
	key("switch", K_SWITCH)
	end_h;

	table(s_tab, 2)
	sub_tab(s5_tab, 5, s5_n),
	sub_tab(s6_tab, 6, s6_n),
	end_tab;

/* ----- T TABLE ----- */
#define t7_n 1
	header(t7_tab, t7_n)
	key("typedef", K_TYPEDEF)
	end_h;

	table(t_tab, 1)
	sub_tab(t7_tab, 7, t7_n),
	end_tab;

/* ----- U TABLE ----- */
#define u5_n 1
	header(u5_tab, u5_n)
	key("union", K_UNION)
	end_h;

#define u8_n 1
	header(u8_tab, u8_n)
	key("unsigned", K_UNSIGNED)
	end_h;

	table(u_tab, 2)
	sub_tab(u5_tab, 5, u5_n),
	sub_tab(u8_tab, 8, u8_n),
	end_tab;

/* ----- V TABLE ----- */
#define v4_n 1
	header(v4_tab, v4_n)
	key("void", K_VOID)
	end_h;

#define v8_n 1
	header(v8_tab, v8_n)
	key("volatile", K_VOLATILE)
	end_h;

	table(v_tab, 2)
	sub_tab(v4_tab, 4, v4_n),
	sub_tab(v8_tab, 8, v8_n),
	end_tab;

/* ----- W TABLE ----- */
#define w5_n 1
	header(w5_tab, w5_n)
	key("while", K_WHILE)
	end_h;

	table(w_tab, 1)
	sub_tab(w5_tab, 5, w5_n),
	end_tab;

/*
	Indirect reserved word table
	WARNING:  this used to be offset from 'A' instead of 'a'.
*/
tab_node * rwx_tab [] = {

	a_tab,
	b_tab,
	c_tab,
	d_tab,
	e_tab,
	f_tab,
	g_tab,
	0L,		/* h */
	i_tab,
	0L,		/* j */
	0L,		/* k */
	l_tab,
	0L,		/* m */
	0L,		/* n */
	0L,		/* o */
	0L,		/* p */
	0L,		/* q */
	r_tab,
	s_tab,
	t_tab,
	u_tab,
	v_tab,
	w_tab,
	0L,		/* x */
	0L,		/* y */
	0L		/* z */
};

/*
	See if name[] contains a reserved word.
	Return the token or NULL_TOK if not found.
	
	This routine should be very fast.
*/
en_tokens
is_reserved(reg char *name, int length)
{
	FTAG("-is_reserved");
	
	reg char			c;
	reg tab_node   *	p;
	reg entry_node *	e;
	reg int 			i, n;

	// TRACEP(ftag, es("name: "); es(name); es(", length: "); eint(length); enl());

	c = *name;
	
	if (c < 'a' || c > 'z') {
		goto not_found;
	}

	/* Search for the proper subtable. */
	for (p = rwx_tab [c - 'a']; ; p++) {

		// TRACEP(ftag, es("p = "); eptr(p); enl());
	
		if (p == NULL || p -> t_entry == NULL) {
			goto not_found;
		}

		// TRACEP(ftag, es("p -> t_strlen = "); eint(p -> t_strlen); enl());

		if (p -> t_strlen > length) {
			goto not_found;
		}
		else if (p -> t_strlen == length) {
			break;
		}
	}

	/* A subtable with the right length has been found. */
	n = p  -> t_entries;
	for (i = 0, e = p  -> t_entry; i < n; e++, i++) {

		// TRACEP(ftag, es("compare: "); es(name); ecs(); es(e -> e_string); enl());

		if (str_eq(name, e -> e_string)) {

			/* Found. */
			token = ((int) (e -> e_arg1)) & 0xff;
			
			// TRACEP(ftag, eret(); eint(token); enl());
			return token;
		}
	}

not_found:
	// TRACEP(ftag, eret(); es("NULL_TOK\n"));
	return NULL_TOK;
}

/*
	Return TRUE if id is a current name of any Sherlock macro.

	This code is slow, but it is called only by outer_expand after
	an actual macro is seen, so its speed is not important.

	The only effect of the SL_DISABLE macro is
	to cause this routine to return TRUE,
	thereby disabling generation of Sherlock macros.
	
	The FTAG macro also causes this routine to return TRUE,
	and in addition, FTAG causes the code to use ftag rather than
	the function name.
*/
bool
is_sherlock(char *id)
{
	FTAG("is_sherlock");

	reg int i = 0;
	reg char * s = NULL;
	bool result = FALSE;
	
	STAT(ftag);
	
	while (s = mn_tab[i]) {
		if (str_eq(id, s)) {
			result = TRUE;
			break;
		}
		++i;
	}
	TRACEP(ftag, es(id); eblank(); eret(); ebool(result); enl());
	return result;
}

