/*
	SPP: header file for token processing.
	
	source:  SPPtok.h
	started: February 14, 1991.
	version: March 16, 1996.
*/

#ifndef SPPtok_h_
#define SPPtok_h_

#pragma once

#include "SPPenum.h"

/*
	Globals describing tokens.

	The code assumes that no token or string will ever be longer than
	MAX_SYMBOL.  If that ever is not so the program may crash.  Thus,
	MAX_SYMBOL should be very large -- say 1000 or more.
*/
#define MAX_SYMBOL	600		/* Max len of symbols/strings.		*/

/*
	Define possible values for ch.
	
	Flags are found in macro replacement texts.
	These flags MUST NOT BE NEGATIVE, so as to fit in a proper char.
*/

#define ARG_FLAG	1
#define POUND_FLAG	2
#define CONCAT_FLAG 3
#define EORT		4
#define EXPAND_OFF	5
#define ARG_OFFSET	'0'

extern char	ch;						/* The NEXT character.					*/
extern en_tokens token;				/* The current token.					*/
extern int	t_subtype;				/* The subtype of token class.			*/
extern char	t_symbol [MAX_SYMBOL];	/* The spelling of the token.			*/
extern int	t_length;				/* The length of the token (in the text)*/
extern long	t_value;				/* Value of integer constants.			*/

extern int	last_tok;		/* Previous token.					*/
extern int	last_ch;		/* Last char.						*/
extern bool	first_tok;		/* True if token is first in line.	*/
extern bool	con_flag;		/* True if in PP constant expr.		*/

/*
	Function prototypes.
*/
void		t_comment		(void);
void		skip_bl			(void);
void		skip_comment	(void);
void		skip_1line		(void);
void		skip_past		(void);
void		skip_pp			(void);
void		skip_ws			(bool nl_flag);
void		t_id			(char *buffer, int max_length);
en_tokens	t_number		(bool dot_flag);
void		t_string		(char *out, int max_out, bool copy_flag);
void		tok_init_again	(void);

int			in_string	(char *in, char *out, int max_out);

/*
	Macros describing tokens and characters.
*/

#define ishex(c) (isdigit(c) || (c >= 'A' && (c <= 'F' || (c>='a' && c<='f')))
#define isid1(c) (isalpha(c) || c == '_')
#define isid2(c) (isalpha(c) || isdigit(c) || c == '_')
#define is(tok) (token == tok)

#define TRACETOK(name) TRACEPB(name, es("tok: "); es_tok(); enl())
#define TRACECH(name)  TRACEPB(name, es("ch: "); es_ch(ch); enl())

#endif /* SPPtok_h_ */
