/*
	SPP: Program termination module.

	Source:  SPPend.c
	Started: November 30, 1990
	Version: March 16, 1996.
*/

#include "SPP.h"

#include "SPPend.h"
#include "SPPenv.h"

#include <stdlib.h>

#include <LIBlog.h>
#include <mac_gui.h>

/*
	Forward declarations of internal routines.
*/
static void	end_wait_for_exit	(char * message);
static void	end_out				(char * kind, char * message);

/*
	Define internal variables.
*/
static bool end_abort_flag = FALSE;

/*
	A signigicant system error has occured.
	
	Wait for the user to choose quit so that the error message is sure
	to be seen, then close all files and exit.
*/
void
end_abort(void)
{
	SL_DISABLE();
	
	/* Make sure we call end_wait_for_exit at most once. */
	if (!end_abort_flag) {
		end_abort_flag = TRUE;
		end_wait_for_exit("\nChoose quit (or command q) immediately!!.\n");
	}
	exit(0);
}

/*
	The program is finished:  Exit after waiting for quit or a keystroke.
*/
void
end_done(void)
{
	/* Make sure we call end_wait_for_exit at most once. */
	if (!end_abort_flag) {
		end_wait_for_exit("\nChoose quit (or command q) to exit.\n");
		end_abort_flag = TRUE;
	}
	
	log_close();
	exit(0);
}

/*
	Quit has been chosen: close all files and exit immediately.
*/
void
end_quit(void)
{
	ecnl();
	es("\nSPP terminated by user.\n");
	
	/* Make sure we call end_wait_for_exit at most once. */
	if (!end_abort_flag) {
		end_abort_flag = TRUE;
	}
	
	log_close();
	exit(0);
}

/*
	There has been an argument error.  Wait for the user to choose quit.
*/
void
end_usage(void)
{
	/* Make sure we call end_wait_for_exit at most once. */
	if (!end_abort_flag) {
		end_abort_flag = TRUE;
		end_wait_for_exit("\nChoose quit (or command q) now.\n");
	}
	
	log_close();

	exit(0);
}

/*
	Wait for the user to choose quit or hit a key.
*/
static void
end_wait_for_exit(char * message)
{

#if ENV_TYPE==ENV_LSC_DIALOG || ENV_TYPE==ENV_LSC_ARGV

	es(message);

	/* Wait for quit to be chosen. */
	while(w_applEvent(0) == TRUE) {
		;
	}

#else

	#if 0  /* No need to wait for a key-strike in MPW or DOS */
	es("Hit any key to continue...\n");
	getchar();
	#endif

#endif /* defined(ENV_MAC) && defined(SHERLOCK) */

}
