/*
	SPP: Application-specific srror handling library.
	
	Source:  SPPerr.c
	Started: November 30, 1990
	Version: April 6, 1996.
*/

#include "spp.h"

#include "objlib.h"

#include "SPPend.h"
#include "SPPenv.h"
#include "SPPerr.h"
#include "SPPobj.h"
#include "SPPstr.h"

#include <LIBcvt.h>
#include <LIBend.h>

#include <Dialogs.h>

// Private globals.

#define ERR_MAX_FILE_NAME 100

long err_line = 0;			/* Current line number. */
bool err_verbose = FALSE;	/* Verbose flag. */

// Private prototypes

static char * err_short_fn	(char * fname);

/*
	Define internal variables.
*/
static bool		err_fatal_ok = FALSE;	/* TRUE: fatal does not abort	*/
static char *	err_fname = NULL;		/* Current file name.		*/
static char *	err_func_name = NULL;	/* Current function name.	*/
static long		err_count = 0;			/* Error counts...			*/
static long		err_help = 0;
static long		err_warning = 0;
static bool 	err_abort_flag = FALSE;

/*
	Initialize all the internal variables declared above.
*/
void
err_init_again(void)
{
	err_func_name	= NULL;
	err_line		= 0;

	err_count		= 0;
	err_help		= 0;
	err_warning		= 0;
}

/*
	Set the file name used in error message.
	Make sure that it will always be available.
*/

void
err_begin_file (char * file_name)
{
	SL_DISABLE();

	/* Allocate the buffer if needed. */
	if (err_fname == NULL) {
		err_fname = obj_alloc(ERR_MAX_FILE_NAME, OBJ_LIB_FILE_NAME_BUF);
	}
	str_ncpy(err_fname, file_name, ERR_MAX_FILE_NAME);
}

void
err_begin_func (char * func_name)
{
	SL_DISABLE();
	
	if (err_func_name == NULL) {
		err_func_name = obj_alloc(ERR_MAX_FILE_NAME, OBJ_LIB_FUNC_NAME_BUF);
	}
	str_ncpy(err_func_name, func_name, ERR_MAX_FILE_NAME);

	if (err_verbose) {
		es(err_func_name); enl();
	}
}

/*
	Return the name of the current file or "(no file)".
*/
char *
err_file_name(void)
{
	SL_DISABLE();
	
	if (err_fname == NULL) {
		return "(no file)";
	}
	else {
		return err_short_fn(err_fname);
	}
}

/*
	Given a static string, return the short file name of the string.
*/
char *
err_short_fn(char * file_name)
{
	reg int i, n;
	
#ifdef ENV_MAC
		
	/* Scan backwards looking for a colon. */
	n = strlen(file_name)-2;
	for (i = n; i >= 0; i--) {
		if (file_name[i] == ':') {
			return file_name + i + 1;
		}
	}
	return file_name;
	
#else
	return file_name;
#endif

}


#define F_WIDTH 10

static void err_common(char * title);

static void
err_common(char * title)
{
	
	char * name = ((err_func_name == NULL) ? "" : err_func_name);
	int    pad  = max(0,F_WIDTH - strlen(name));
	
	ecnl();
	epads(title,-8);
	es(" line: ");
	epadlong(err_line,4); eblank();
	es(err_short_fn(err_fname));
	es(" (");
		etrunc(name, F_WIDTH);
	es(") ");
	eblanks(pad);
}

bool
err_begin(void)
{
	err_common("error");
	err_count++;
	return TRUE;
}

void
err_context(void)
{
	ecnl();
}

static bool in_fatal = FALSE;

void
fatal_begin(void)
{
	ecnls(2);
	es("Oh Dear...There is something wrong me\n");
	es("in: "); es(current_ftag);
	es(" called from "); es(caller_ftag);
	es(" called from "); es(caller2_ftag);
	enl();

	if (in_fatal) {
		es("Encountered a second fatal error...aborting now!\n");
		end_abort();
	}
	in_fatal = TRUE;
}

void
fatal_end(void)
{
	ecnls(2);
	end_abort();
}

void
note_begin(void)
{
	err_common("note");
}

void
warning_begin(void)
{
	err_common("warning");
}
