/*
	Header for application-specific io library.
	
	source:  SPPio.h
	started: November 29, 1990
	version: March 16, 1996.
*/

#ifndef SPPio_h_
#define SPPio_h_

#pragma once

#include "SPPlst.h"

#undef	IO_DEBUG
#define	IO_DEBUG

/*
	Define the size of stream buffers.
*/
#define IO_BUF_SIZE 1024

#define END_FILE 0x1a

/*
	Define the status of a file.
*/
typedef enum { IO_NULL,
	IO_CLOSED, IO_INPUT, IO_OUTPUT, IO_EOF
} io_stat;


/*
	Define the format of path names and file structs.
	
	The vRefNum field is used only on the Macintosh.
*/

#define IO_NULL_VREFNUM 0

#ifndef LIB_IO_TYPEDEFS
typedef struct io_file_struct	io_file;
#endif

typedef struct io_path_struct	io_path;
typedef struct io_strm_struct	io_strm;

struct io_path_struct {
	LIST;
	char *		io_path_name;
};

struct io_file_struct {
	LIST;
	io_stat		io_file_stat;	/* Status of file		*/
	io_path *	io_file_path;	/* File path.			*/
	FILE *		io_file_handle;	/* File handle.			*/ // 3/15/96
	int			io_file_line;	/* Line number.			*/
	int			io_file_lastc;	/* Restart character.	*/
};

struct io_strm_struct {
	LIST;
	io_file *	io_strm_file;	/* File for this stream.		*/
	
	
	char *		io_strm_buffer;	/* File buffer					*/
	int			io_strm_size;	/* File buffer size (bytes)		*/
	char *		io_strm_bufp;	/* Pointer into file buffer		*/
	int			io_strm_bufc;	/* Characters in buffer			*/
};

/*
	Declare global input stream variables.
*/
extern int		 io_in_strm_level;	/* Level of input stream.		*/
extern io_strm * io_strm_in_list;	/* List of open input streams.	*/
extern io_strm * io_strm_out_list;	/* List of open output streams.	*/

/*
	Function prototypes for path routines.
*/
void		io_close_path		(io_path * VAR path);
io_path *	io_new_path_struct	(char * path_name);

/*
	Function prototypes for file routines.
*/
void		io_close		(io_file * file);
io_file *	io_create		(io_path * path, bool cant_flag);
bool		io_is_open		(io_file * file);
long		io_length		(io_file * file);
io_file *	io_open			(io_path * path, bool cant_flag);
long		io_read			(io_file * file, char * buffer, long n);
char *		io_read_all		(io_path * path);
long		io_write		(io_file * file, char * buffer, long n);

/*
	Function prototypes for stream routines.
*/
void		io_close_strm	(io_strm * stream);
io_strm *	io_create_strm	(io_path * path, bool cant_flag);
char		io_getc			(io_strm * file);
io_strm *	io_new_strm		(io_file * file, int buf_size);
io_strm *	io_open_strm	(io_path * path, bool cant_flag);
void		io_putc			(char c);
void		io_puts			(char * s);
void		io_show_all		(bool warn_flag);

/*
	Function prototypes for dumping routines.
*/
void		io_show_file	(io_file * file);
void		io_show_path	(io_path * path);
void		io_show_strm	(io_strm * stream);

#endif /* SPPio_h_ */