/*
	SPP: Header for list library.
	
	All list elements must be objects allocated by obj_alloc or obj_new.
	
	source:  SPPlst.h
	started: December 14, 1990.
	version: March 16, 1996.
*/

#ifndef SPPlst_h_
#define SPPlst_h_

#pragma once

/*
	Declare the elements of singly-linked lists.
*/
#define LIST\
	lst_node *	lst_nxt		/* Pointer to next object.	*/
	
typedef struct lst_struct  lst_node;

struct lst_struct {
	LIST;
};

/*
	Declare the elements of doubly-linked lists.
*/
#define LIST2\
	lst2_node *	lst2_nxt;	/* Pointer to next object.		*/	\
	lst2_node *	lst2_back	/* Pointer to previous object.	*/

typedef struct lst2_struct lst2_node;

struct lst2_struct {
	LIST2;
};

/*
	Declare the basic routines.
	
	lst_safe_every and lst2_safe_every will work even if node p is deleted.
*/
#define lst_every(c, p,list)\
 for(p=(c *)*list;p!=NULL;p=(c *)(((lst_node *)p)->lst_nxt))
 
#define lst_safe_every(c, p, p2, list)\
 for(p=(c *)*list,p2=((p==NULL)?NULL:(c *)((lst_node *)p->lst_nxt));\
 	p!=NULL;p=(c *)p2,p2=((p==NULL)?NULL:(c *)((lst_node *)p->lst_nxt)))
 
#define lst2_every(c, p,list)\
 if(*list!=NULL)\
 	for(p=(c *)((*list)->lst2_nxt);p!=NULL;p=(c *)(((lst2_node *)p)->lst2_nxt))
 
#define lst2_safe_every(c, p, p2, list)\
 if(*list!=NULL)\
 	for(p=(c *)((*list)->lst2_nxt),p2=((p==NULL)?NULL:(c *)((lst2_node *)p->lst2_nxt));\
 		p!=NULL;p=(c *)p2,p2=((p==NULL)?NULL:(c *)((lst2_node *)p->lst2_nxt)))

#define	lst_add(n,l)		lst_add_	((lst_node *) n, (lst_node **) l)
#define	lst_append(n,l)		lst_append_	((lst_node *) n, (lst_node **) l)
#define	lst_free_all(l)		lst_free_all_((lst_node **) l)
#define	lst_member(n,l)		lst_member_	((lst_node *) n, (lst_node **) l)
#define lst_next(l)			lst_next_	((lst_node **) l)
#define	lst_remove(n,l)		lst_remove_ ((lst_node *) n, (lst_node **) l)
#define lst_report(f,l,n)	lst_report_ (f, (lst_node **) l, n);

#define	lst2_add(n,l)		lst2_add_	((lst2_node *) n, (lst2_node **) l)
#define	lst2_append(n,l)	lst2_append_((lst2_node *) n, (lst2_node **) l)
#define	lst2_free_all(l)	lst2_free_all_((lst2_node **) l)
#define	lst2_member(n,l)	lst2_member_((lst2_node *) n, (lst2_node **) l)
#define lst2_next(l)		lst2_next_	((lst_node **) l)
#define	lst2_remove(n)		lst2_remove_ ((lst2_node *) n)
#define lst2_report(f,l,n)	lst_report_ (f, (lst_node **) l, n);

void	lst_add_		(lst_node * node, lst_node ** list);
void	lst_append_		(lst_node * node, lst_node ** list);
void	lst_free_all_	(lst_node ** list);
bool	lst_member_		(lst_node * node, lst_node ** list);
lst_node *	lst_next_	(lst_node ** list);
int		lst_remove_		(lst_node * node, lst_node ** list);
void	lst_report_		(void (*f) (lst_node *), lst_node ** list, char * name);

void	lst2_add_		(lst2_node * node, lst2_node ** list);
void	lst2_append_	(lst2_node * node, lst2_node ** list);
void	lst2_free_all_	(lst2_node ** list);
bool	lst2_member_	(lst2_node * node, lst2_node ** list);
lst2_node *	lst2_next_	(lst2_node ** list);
void	lst2_remove_	(lst2_node * node);
void	lst2_report_	(void (*f) (lst2_node *), lst2_node ** list, char * name);

#endif /* SPPlst_h_ */
