/*
	SPP: Object allocation routines.
		
	None of the routines in this file should be called directly.
	Instead, use the macros defined in obj.h.

	source:  SPPobj.c
	started: September 20, 1985
	version: March 16, 1996.
*/

#include "SPP.h"

#include "objlib.h"

#include "SPPend.h"	/* Additional library. */
#include "SPPenv.h"
#include "SPPerr.h"
#include "SPPlst.h"
#include "SPPobj.h"
#include "SPPstr.h"

#ifdef ENV_TURBOC
	#include <alloc.h>
#endif

#ifdef ENV_MSC
	#include <malloc.h>
#endif

/*
	Define fields the object header.
*/

#define M_HEADER_SIZE	4		/* Size of obj_prot field.	*/
#define M_TRAILER_SIZE	8		/* Size of trailer field.	*/
#define M_FILLER		0xA3	/* Fill value for fields.	*/
#define OBJ_CLASS_FIELD 0x0fff	/* Class bits of obj_theType	*/

/*
	Define the hidden "object header" which precedes all objects
	allocated by new_object.
	
	obj_new and obj_alloc return a pointer to the byte *following* this header
	so these fields are *not* accessible outside this class.
*/
typedef struct obj_struct obj_header;

static struct obj_struct {
	unsigned	obj_theType;				/* Object type.			*/
	
#if OBJ_KIND == OBJ_DEBUG
	char 		obj_prot[M_HEADER_SIZE];	/* Header protection.	*/
#endif

};

/*	
	The order of elements in this array must match the order of
	elements in obj_enum in obj.h.
*/

static char * m_tag2name [] = {

	/* OBJ_BAD				*/	"bad object",

	/* OBJ_LIB_MNODE		*/	"list:lib m_node",
	/* OBJ_LIB_FILE_STRUCT	*/	"list:lib file node",
	/* OBJ_LIB_PATH_STRUCT	*/	"list:lib path node",
	/* OBJ_LIB_IOSTRM_STRUCT*/	"list:lib io strm node",
	
	/* OBJ_MST_STRUCT		*/	"list:macro node",
	/* OBJ_UNDEF_STRUCT		*/	"list:macro unode",
	/* OBJ_ST_LIST_STRUCT	*/	"list:st list",
	/* OBJ_ST_LEVEL_STRUCT	*/	"list:st level",
	
	/* OBJ_LIB_LST2_HEAD	*/	"lst2:lib head",
	/* OBJ_LIB_STRM_STRUCT	*/	"lst2:lib strm node",
	/* OBJ_ST_NODE_STRUCT	*/	"lst2:st typedef node",

	/* End of list classes. */
	
	/* OBJ_LIB_ARG_FILE_BUF	*/	"lib io readall buffer",
	/* OBJ_LIB_FILE_NAME_BUF */	"lib err file name buffer",
	/* OBJ_LIB_FUNC_NAME_BUF */	"lib err func name buffer",
	
	/* OBJ_ARGV_BUF			*/	"argv buffer",
	/* OBJ_ARG_DEF_BUF		*/	"arg_def_buf",
	/* OBJ_ARG_UNDEF_BUF	*/	"arg_undef_buf",
	/* OBJ_LOC_DEF_BUF		*/	"locDefineBuf",
	/* OBJ_LOC_UNDEF_BUF	*/	"locUndefBuf",
	
	/* OBJ_LIB_PATH_NAME	*/	"path name buffer",
	/* OBJ_LIB_STRM_BUFFER	*/	"io strm buffer",
	/* OBJ_STRM_BUFFER		*/	"strm buffer",
	
	/* OBJ_EXPAND_OLD_ATEXT	*/	"expand: old_atext",
	/* OBJ_EXPAND_NEW_ATEXT	*/	"expand: new_atext",
	/* OBJ_RESCAN_LOCAL_IN	*/	"rescan: local_in",
	/* OBJ_RESCAN_EXPAND_BUF */	"rescan: expand_buf",
	/* OBJ_RESCAN_ID_BUF	*/	"rescan: id_buf",
	
	/* OBJ_SYN_STRING		*/	"synonym string",
	/* OBJ_PATH_STRING		*/	"path name string",
	/* OBJ_TYPDEF_STRING	*/	"typedef string",
	/* OBJ_PARAM_STRING		*/	"param string",
	
	/* OBJ_MACRO_STRING		*/	"macro name string",
	/* OBJ_MACRO_RTEXT		*/	"macro rtext string",
	/* OBJ_MACRO_LINE		*/	"__line__ string",
	/* OBJ_MACRO_FILE		*/	"__file__ string",
	/* OBJ_MACRO_TIME		*/	"__time__ string",
	/* OBJ_MACRO_DATE		*/	"__date__ string",

	/* OBJ_LAST_INDEX		*/	"bad object (last index)"
};

/*
	Define the format of a node descriptor.
*/

typedef struct m_struct m_node;

static struct m_struct {
	LIST;
	int			m_type;		/* Type of memory.						*/
	char	*	m_allocp;	/* Ptr to allocated memory.				*/
	int			m_size;		/* Logical size of allocated memory.	*/
};

/*
	Define the list of allocated blocks.
	
	obj_alloc adds items to the list and obj_free removes them.
*/
static m_node * 	m_alloc_list = NULL;

/*
	Defind statistics arrays:  one array element for each class.
*/
static long		obj_tot		[OBJ_LAST_INDEX];
static long		obj_net		[OBJ_LAST_INDEX];
static long		obj_max		[OBJ_LAST_INDEX];
static long		obj_nalloc	[OBJ_LAST_INDEX];
static long		obj_nfree	[OBJ_LAST_INDEX];
static char *	obj_statTag	[OBJ_LAST_INDEX];

static bool obj_init_flag = FALSE;	/* TRUE: statistics have been initialized.	*/

/*
	Declare internal routines used only during debugging.
*/
static void		m_checkfill	(char * p, int n, bool trail_flag,
							 char * tag1, char * tag2);
static void		m_dump1		(m_node * p, bool all_flag);
static void		m_dumpfill	(char *p, int n, char * tag);
static char *	m_prtype	(int type);

/*
	Object allocator:  Allocate n bytes for the "logical" object
	preceded by the object header and possibly followed by the protection field.
*/
void *
m_alloc(int n, int type)
{
	FTAG("m_alloc");

	reg char *			p;
	reg char *			q;
	reg obj_header *	o;
	reg int				i;

	/* Align the request. */
	if (n & 1) {
		n++;
	}
	
#if OBJ_KIND == OBJ_DEBUG

	if (type & M_PROT) {
		p = calloc((size_t) 1,
			(size_t) (n + sizeof(obj_header) + M_TRAILER_SIZE) );
	}
	else {
		p = calloc((size_t) 1, (size_t) (n + sizeof(obj_header)) );
	}
	
#else

	p = calloc((size_t) 1, (size_t) (n + sizeof(obj_header)) );
	
#endif /* OBJ_KIND == OBJ_DEBUG */
	
	if (p == NULL) {
		ecnls(2); es("sorry, can not allocate: "); elong(n); es(" bytes."); enl();
		obj_stat(ftag);
		obj_show(TRUE, ftag);
		end_abort();
	}
	
	o = (obj_header *) p;
	p += sizeof(obj_header);
	
	/* Fill in the hidden object type. */
	o -> obj_theType = type;

#if OBJ_KIND == OBJ_DEBUG

	/* Fill in the header. */
	for(i = 0; i < M_HEADER_SIZE; i++) {
		o -> obj_prot[i] = M_FILLER;
	}
	
	/* Fill in the trailer. */
	if (type & M_PROT) {
		for (q = p + n, i = 0; i < M_TRAILER_SIZE; i++) {
			*q++ = M_FILLER;
		}
	}
	
#endif /* OBJ_KIND == OBJ_DEBUG */
	
	return p;
}

void *
m_alloc_db (int n, int type)
{
	FTAG("m_alloc_db");

	reg char * 		p;			/* External object.	*/
	reg m_node * 	node_p;
	reg int 		i;
	reg int			class;

	TRACEPB(ftag,
		es("n: ");		epadint(n,4);
		es(" type:");	epads(m_prtype(type), 5);
		es(":");		epadint(type & OBJ_CLASS_FIELD, 4);
		es(" ");		epads(m_tagType(type),-25));
		
	/* Check the class. */
	class = (type & ~M_MASK);
	if (class <= 0 || class >= OBJ_LAST_INDEX) {
		fatal(es("m_alloc_db: bad class: "); eint(class); enl());
	}
		
	/* Initialize if needed. */
	if (!obj_init_flag) {
		
		for (i = 0; i < OBJ_LAST_INDEX; i++) {
			obj_tot		[i] = 0L;
			obj_net		[i] = 0L;
			obj_max		[i] = 0L;
			obj_statTag	[i] = NULL;
		}
		obj_init_flag = TRUE;
	}
	
	/* Get the memory. */
	p = m_alloc(n, type);
	
	/* Allocate a new descriptor node. */
	node_p = CAST(m_node *) m_alloc(sizeof(m_node), OBJ_LIB_MNODE);
	
	/* Fill in the node and put it on the list of allocated blocks. */
	node_p -> m_size = n;
	node_p -> m_allocp = p;
	node_p -> m_type = type;
	
	lst_add(node_p, &m_alloc_list);

	/* Update cumulative statistics (held in the OBJ_BAD slot). */
	obj_nalloc	[0]	++;
	obj_tot		[0] += (long) n;
	obj_net		[0] += (long) n;
	obj_max		[0]  = max(obj_max[0], obj_net[0]);

	/* Update class statistics. */
	obj_nalloc	[class]	++;
	obj_tot		[class] += (long) n;
	obj_net		[class] += (long) n;
	obj_max		[class]  = max(obj_max[class], obj_net[class]);
	
	/* Check the list for consistency. */
	TRACE("m_alloc_db_check", m_checkMem("m_alloc_db"));
	
	TRACEX("m_alloc_db", eret(); eptr(p); enl());
	return p;
}

/*
	Check the type of object of object o.
*/
void
m_check	(void * obj, int class)
{
	SL_DISABLE();
	
	if (m_type(obj) != class) {
		fatal(
			es("m_check: type: "); es(m_tag(obj));
			es(" is not "); es(m_tagType(class)); enl());
	}
}

void
m_check2 (void * obj, int class1, int class2)
{
	SL_DISABLE();
	
	if (m_type(obj) < class1 || m_type(obj) > class2) {
		fatal(
			es("m_check: type: "); es(m_tag(obj));
			es(" is not between "); es(m_tagType(class1));
			es(" and "); es(m_tagType(class2)); enl());
	}
}

void
m_check_db (void * obj, int class, char *tag)
{
	SL_DISABLE();
		
	if (m_type(obj) != class) {
		fatal(
			es("m_check:"); es(tag);
			es(" type: "); es(m_tag(obj));
			es(" is not "); es(m_tagType(class)); enl());
	}
}
	
void
m_check2_db (void * obj, int class1, int class2, char *tag)
{
	SL_DISABLE();
	
	if (m_type(obj) < class1 || m_type(obj) > class2) {
		fatal(
			es("m_check:"); es(tag);
			es(" type: "); es(m_tag(obj));
			es(" is not between "); es(m_tagType(class1));
			es(" and "); es(m_tagType(class2)); enl());
	}
}

/*
	Check the header and trailer areas.
	
	n is the size of the block *not* counting the header and trailer.
	p is a pointer to the memory block, not the header.
*/
static void
m_checkfill(char * p, int n, bool trail_flag, char * tag1, char * tag2)
{
	obj_header *	o;
	char * q;
	int i, m;
	unsigned int val;
		
	TRACE("no_checkfill", return;);
	
	m = ( (n & 1) ? (n + 1) : n);
	o = (obj_header *) (p - sizeof(obj_header));
	
#if OBJ_KIND == OBJ_DEBUG

	/* Check the header. */
	for(i = 0; i < M_HEADER_SIZE; i++) {
		val = (o -> obj_prot[i]) & 0xff;
		if (val != M_FILLER) {
			m_dumpfill(&(o->obj_prot[0]), M_HEADER_SIZE, "Bad header");
			fatal(
				es("m_checkfill: ");
				es(tag1); es(":"); es(tag2);
				es(" p: "); eptr(p);
				es(" n: "); eint(n);
				es(" header byte "); eint(i); enl());
		}
	}
#endif
	
	/* Check the trailer. */
	if (trail_flag) {
		q = p + m;
		for (i = 0; i < M_TRAILER_SIZE; i++, q++) {
			val = (*q) & 0xff;
			if (val != M_FILLER) {
				m_dumpfill(p + m, M_TRAILER_SIZE, "Bad trailer");
				fatal(
				es("m_checkfill: ");
				es(tag1); es(":"); es(tag2);
				es(" p: "); eptr(p);
				es(" n: "); eint(n);
				es(" trailer byte "); eint(i); enl());
			}
		}
	}
}

/*
	Check the list of all know objects for consistency.
*/
void
m_checkMem(char * tag)
{
	reg m_node *	node_p;
	reg int			i;
	reg int			type;
	reg int			class;
	long			sum [OBJ_LAST_INDEX]
	
	SL_DISABLE();
	
	for (i = 0; i < OBJ_LAST_INDEX; i++) {
		sum[i] = 0;
	}
	
	/*
		Sum the bytes of each class.
	*/
	lst_every(m_node, node_p, &m_alloc_list) {
	
		type = node_p -> m_type;

		m_checkfill(node_p -> m_allocp, node_p -> m_size, 
			(type & M_PROT), m_tag(node_p -> m_allocp), tag);
		
		class = (type & ~M_MASK);
		if (class <= 0 || class >= OBJ_LAST_INDEX) {
			fatal(es("m_checkMem:  bad class"));
		}
		sum [class] += node_p -> m_size;
	}
		
	/* Verify the net totals. */
	for (i = 1; i < OBJ_LAST_INDEX; i++) {
		if (sum[i] != obj_net[i]) {
			fatal(
				es("m_checkMem:"); es(tag);
				es(" sum["); eint(i); es("]: "); eint(sum[i]);
				es("d != net["); eint(i); es("] "); eint(obj_net[i]); enl());
		}
	}
}

/*
	Make sure obj is a valid object.
*/
void
m_checkObj(void * obj, char * tag)
{
	FTAG("m_checkObj");

	reg m_node *	node_p;
	reg int			type;
	reg int			class;
	
	lst_every(m_node, node_p, &m_alloc_list) {
		if (node_p -> m_allocp == obj) {
		
			/* Found. */
			type = node_p -> m_type;
			class = (type & ~M_MASK);
			
			if (class <= 0 || class >= OBJ_LAST_INDEX) {
				fatal(es("m_checkObj:  bad class: "); eint(class); enl());
			}
			
			/* Check the header and trailer bytes. */
			m_checkfill(node_p -> m_allocp, node_p -> m_size,
				(type & M_PROT), m_tag(node_p -> m_allocp), "m_checkObj");

			return;
		}
	}
	
	/* Not found. */
	obj_stat(ftag);
	obj_show(TRUE, ftag);
	fatal(
		es(ftag); es(": "); es(tag);
		es(": unallocated pointer: "); eptr(obj); enl());
}

/*
	Dump the value of a single m_node.
*/
static void
m_dump1(m_node * p, bool all_flag)
{
	reg int type = p -> m_type;
	
	if (!all_flag && (type & M_NODUMP)) {
		return;
	}
	
	es("addr: ");	eptr(p -> m_allocp);
	es(" size: ");	epadint(p -> m_size, 4);
	es(" type:");	epads(m_prtype(type), 5);
	es(":");		epadint(type & OBJ_CLASS_FIELD, 4);
	es(":");		epads(m_tag(p->m_allocp), -25);

	/* Print the value of the string. */
	if (type & M_STRING) {
			
		if (strlen(p -> m_allocp) > 10) {
			enl(); es("\t");
		}
		else {
			eblank();
		}
		es("string: "); es(p -> m_allocp);
	}
	ecnl();
}

/*
	Print the fill area.
*/
static void
m_dumpfill(char *p, int n, char * tag)
{
	int i;
	unsigned int val;
	
	SL_DISABLE();
	
	ecnl(); es(tag); enl();

	for (i = 0; i < n; i++) {
		val = (p[i]) & 0xff;
		if (val == M_FILLER) {
			es("<FILL>");
		}
		else {
			echar(p[i]);
			es("<"); ehex(p[i]); es(">");
		}
	}
	ecnl();
}

/*
	Free memory allocated by m_alloc().
	
	Different versions are needed since they have different
	calling sequences.
*/
void
m_free(void * p2)
{
	reg char * p = (char *) p2;
	
	SL_DISABLE();
	
	if (p == NULL) {
		fatal(es("m_free: trying to free a NULL pointer"));
	}
	free(p - sizeof(obj_header));
}

/*
	Warning:  trying to free a string will cause an error.
*/
void
m_free_db(void * p)
{
	FTAG("m_free_db");
	FTAGV("-m_free_db_v");
	
	reg m_node *	node_p;
	reg int			type;
	reg char *		save_p;
	reg int			class;
	
	SL_DISABLE();
	
	if (p == NULL) {
		fatal(es("m_free_db: trying to free a NULL pointer"));
	}
	
	lst_every(m_node, node_p, &m_alloc_list) {
	
		if (node_p -> m_allocp == p) {
		
			type = node_p -> m_type;
					
			TRACEP(ftag,
				es("*n: ");		epadint(node_p -> m_size, 4);
				es(" type:");	epads(m_prtype(type), 5);
				es(":");		epadint(type & OBJ_CLASS_FIELD, 4);
				es(" ");		epads(m_tag(p), -25);
				es("p: ");		eptr(p); enl());

			/* Reduce the net bytes allocated. */
			
			class = (type & ~M_MASK);
			if (class <= 0 || class >= OBJ_LAST_INDEX) {
				fatal(es("m_free_db:  bad class: "); eint(class); enl());
			}
			
			obj_nfree	[0]++;
			obj_net 	[0] -= node_p -> m_size;
			obj_nfree	[class]++;
			obj_net 	[class] -= node_p -> m_size;
			
			lst_remove(node_p, &m_alloc_list);
			
			TRACE("m_free_db_check", m_checkMem("m_alloc_db"));
			
			/* Check the header and trailer bytes. */
			m_checkfill(node_p -> m_allocp, node_p -> m_size,
				(type & M_PROT), m_tag(node_p -> m_allocp), "m_free_db");
				
			m_checkfill( (char *) node_p, sizeof(m_node),
				FALSE, m_tag(node_p), "m_free_db (m_node)");

			/*
				Free the m_node before the main block.
				This ought to help free() a bit.
				Careful:  don't use node_p after freeing it.
			*/
			save_p = node_p -> m_allocp;
			TRACEP(ftagv,
				es("Before freeing ");
				eint(node_p -> m_size); es("+");
				eint(sizeof(m_node)); enl());
			
			m_free(node_p);
			m_free(save_p);
	
			TRACEP(ftagv, es("after...\n"));
			
			return;
		}
	}
	
	/* Not found. */
	fatal(
		es("m_free: "); es(m_tag(p));
		es(": unallocated pointer: ");  eptr(p); enl();
		obj_stat(ftag);
		obj_stat(ftag));

}

/*
	Return a string indicating the type of an object.
*/
#define M_PRMAX 20
static char m_prbuf [M_PRMAX];

static char *
m_prtype(int type)
{
	SL_DISABLE();

	m_prbuf[0] = '\0';
	strcat(m_prbuf, (type & M_STRING) ? "S" : "s");
	strcat(m_prbuf, (type & M_PROT)   ? "P" : "p");
	strcat(m_prbuf, (type & M_NOFILL) ? "f" : "F");
	strcat(m_prbuf, (type & M_NODUMP) ? "d" : "D");

	return m_prbuf;
}

/*
	Print report of allocated blocks.
*/
void
m_show(bool all_flag, char * tag)
{
	reg m_node * p;
	
	SL_DISABLE();

	ecnls(2); es(tag); es(": Allocated blocks: \n\n");
	
	lst_every(m_node, p, &m_alloc_list) {
		m_dump1(p, all_flag);
	}
}

void
m_stat(char * tag)
{
	reg int i;

	SL_DISABLE();
	
	ecnls(2); es(tag); es(": Class statistics\n\n");

	for (i = 1; i < OBJ_LAST_INDEX; i++) {
	
		// log_fmt6("%25s:%8ld net:%6ld max:%6ld alloc:%4ld free:%4ld\n",
						epads(m_tagType(i), 25); 
		es(":");		epadlong(obj_tot[i], 8);
		es(" net:");	epadlong(obj_net[i], 6);
		es(" max:");	epadlong(obj_max[i], 6);
		es(" alloc:");	epadlong(obj_nalloc[i], 4);
		es(" free:"); 	epadlong(obj_nfree[i],  4); enl();
	}
					epads("Totals", 25);
	es(":");		epadlong(obj_tot[0], 8);
	eblanks(5); 	epadlong(obj_net[0], 6);
	eblanks(5);		epadlong(obj_max[0], 6);
	eblanks(7);		epadlong(obj_nalloc[0], 4);
	eblanks(6);		epadlong(obj_nfree[0],  4); enl();
			
	es("OBJ_LAST_INDEX: "); eint(OBJ_LAST_INDEX); enl();

	m_checkMem("tag");
}

/*
	Return the type string of an object or type.
*/
char *
m_tag (void * obj)
{
	SL_DISABLE();
	
	return m_tagType(m_type(obj));
}

char *
m_tagType(int type)
{
	reg int class;
	
	SL_DISABLE();
	
	class = (type & ~M_MASK);
	
	if (class < 0 || class >= OBJ_LAST_INDEX) {
		return "m_tag: unknown tag";
	}
	else {
		return m_tag2name[class];
	}
}
	
/*
	Return the type of an object.
*/
int	
m_type (void * obj)
{
	char * 			p;
	obj_header *	h;
	
	SL_DISABLE();
	
	p = CAST(char *) obj;
	h = CAST(obj_header *) (p - sizeof(obj_header));
	return (h -> obj_theType) & OBJ_CLASS_FIELD;
}

/*
	Test this module for blunders.
	Memory allocation stress test.
*/
#define MAX_IT	50
#define MAX_OBJ 10
#define OBJ_SIZE 50

static char *	objects [MAX_OBJ];

void
obj_test (void)
{
	reg int i, j;
	
	/*
		This is a test to see whether opb_free really is freeing memory.
	*/
	
	ecnls(2);
	es("Start of allocation test\n\n");
	
	sl_on("m_*");
	sl_off("m_free_db_v");

	/*
		Start loops at 1 so we do not request 0 bytes.
	*/
	for (i = 1; i < MAX_IT; i++) {
		for (j = 1; j < MAX_OBJ; j++) {
			objects [j] = obj_alloc(OBJ_SIZE * j * i, OBJ_STRM_BUFFER);
		}
		for (j = 1; j < MAX_OBJ; j++) {
			obj_free(objects [j]);
		}
		es("End of iteration "); eint(i); enl();
	}
	
	sl_off("m_*");
	
	es("Allocation test complete\n\n");
}
