/*
	SPP: application-specific memory routines.

	source:  SPPobj.h
	started: November 30, 1990.
	version: March 16, 1996.
*/

#ifndef SPPobj_h_
#define SPPobj_h_

#pragma once

#include <stdlib.h>		/* For malloc */

/*
	Define markers for casts.
*/
#define CAST(a) 		(a)
#define SUPER(a)		(a)		/* Cast to super class of a.	*/
#define SUB_CLASS(a)	(a)		/* Cast to sub class of a.		*/

/*
	Define the object type field returned by the obj_type method.
	
	There are two kinds of partitions on this list:
	
		class partitions:		used by the type checking functions.
		allocation partitions:	used by obj_alloc and obj_free. 
		
	The class partitions are as follows:
	
		subclasses of the list class.
		subclasses of the stream class.
	
	The allocation partitions are as follows:
	
	 memory cleared:  		default: cleared.
	 memory protected:		default: not protected.
	 dumped by m_stat:		default: dumped.
	 where allocated:  	strings v. non-strings.
*/
#define OBJ_LIST_BEGIN	OBJ_LIB_MNODE		/* First list class.	*/
#define OBJ_LIST_END	OBJ_ST_LEVEL_STRUCT	/* Last list class.		*/

#define OBJ_LIST2_BEGIN	OBJ_LIB_LST2_HEAD	/* First list2 class.	*/
#define OBJ_LIST2_END	OBJ_ST_NODE_STRUCT	/* Last list2 class.	*/

#define OBJ_STRM_BEGIN	OBJ_LIB_STRM_STRUCT		/* First stream class.	*/
#define OBJ_STRM_END	OBJ_LIB_STRM_STRUCT		/* Last stream class.	*/

typedef enum {
	OBJ_BAD = 0,			/* Bad object.					*/
	
	/*
		Start of list classes.
		Steams are subclass of the list class.
	*/
	
	OBJ_LIB_MNODE,			/* list:m_node				*/
	OBJ_LIB_FILE_STRUCT,	/* list:file struct			*/
	OBJ_LIB_PATH_STRUCT,	/* list:path struct			*/
	OBJ_LIB_IOSTRM_STRUCT,	/* list:io strm struct		*/
	
	OBJ_MST_STRUCT,			/* list:macro node			*/
	OBJ_UNDEF_STRUCT,		/* list:undef node			*/
	OBJ_ST_LIST_STRUCT,		/* list:st list	node		*/
	OBJ_ST_LEVEL_STRUCT,	/* list:st level node		*/
	
	OBJ_LIB_LST2_HEAD,		/* lst2 head				*/
	OBJ_LIB_STRM_STRUCT,	/* lst2:strm node			*/
	OBJ_ST_NODE_STRUCT,		/* lst2:st node				*/
	
	/* End of list classes.		*/
	
	OBJ_LIB_ARG_FILE_BUF,	/* arg.c: arg_file_buffer		*/
	OBJ_LIB_FILE_NAME_BUF,	/* File name buffer				*/
	OBJ_LIB_FUNC_NAME_BUF,	/* Function name buffer			*/

	OBJ_ARGV_BUF,			/* arg.c: argv buffer			*/
	OBJ_ARG_DEF_BUF,		/* arg_dialog: arg_def_buf		*/
	OBJ_ARG_UNDEF_BUF,		/* arg_dialog: arg_undef_buf	*/
	OBJ_LOC_DEF_BUF,		/* arg_dialog: locDefineBuf		*/
	OBJ_LOC_UNDEF_BUF,		/* arg_dialog: locUndefBuf		*/

	OBJ_LIB_PATH_NAME,		/* io_new_path name				*/
	OBJ_LIB_STRM_BUFFER,	/* io_new_strm buffer			*/
	OBJ_STRM_BUFFER,		/* strm buffer					*/

	OBJ_EXPAND_OLD_ATEXT,	/* expand: old_atext			*/
	OBJ_EXPAND_NEW_ATEXT,	/* expand: new_atext			*/
	OBJ_RESCAN_LOCAL_IN,	/* rescan: local_in				*/
	OBJ_RESCAN_EXPAND_BUF,	/* rescan: expand_buf			*/
	OBJ_RESCAN_ID_BUF,		/* rescan: id_buf				*/
	
	OBJ_SYN_STRING,			/* synonym string				*/
	OBJ_PATH_STRING,		/* path name string				*/
	OBJ_TYPDEF_STRING,		/* typedef string				*/
	OBJ_PARAM_STRING,		/* parameter string				*/
	
	OBJ_MACRO_STRING,		/* macro name string			*/
	OBJ_MACRO_RTEXT,		/* macro replacement text		*/
	OBJ_MACRO_LINE,			/* __line__						*/
	OBJ_MACRO_FILE,			/* __file__						*/
	OBJ_MACRO_TIME,			/* __time__						*/
	OBJ_MACRO_DATE,			/* __date__						*/
	
	OBJ_LAST_INDEX,			/* Last valid type + 1			*/
	
	/*
	The allocation partitions are indicated by a bit field. 
	*/
	
	M_STRING	= 0x1000,	/* Allocate in string space.		*/
	M_PROT		= 0x2000,	/* Add trailing protection bytes.	*/
	M_NODUMP	= 0x4000,	/* Exclude from memory dumps.		*/
	M_NOFILL	= 0x8000,	/* Do not clear memory.				*/
	M_MASK		= 0xf000	/* Reserved allocation bits.		*/

} obj_enum;

/*
	The OBJ_KIND selects how objects are represented.
	
	OBJ_NO_HEADER:  Don't use objects at all.
	
	No object header is used, the obj_type and obj_check macros do
	nothing, and obj_alloc and obj_free expand to malloc and free.
		
	OBJ_NORMAL:		Use "production" objects.
	
	All objects are preceded by a 2-byte header, the obj_type macro is
	functional, the obj_check macros are not functional, obj_alloc and
	obj_free expand to m_alloc and m_free and tags are not functional.
	
	OBJ_DEBUG:		Use "debugging" objects.
	
	All objects are preceded by an expanded header, including several
	"protection bytes" in the object header.  The obj_type and
	obj_check macros are functional and obj_alloc and obj_free expand to
	m_alloc_db and m_free_db.  Tags are functional.
*/
#define OBJ_NO_HEADER	1		/* Don't use objects at all.	*/
#define OBJ_NORMAL		2		/* Use "production" objects.	*/
#define OBJ_DEBUG		3		/* Use "debugging" objects.		*/

#define OBJ_KIND		OBJ_DEBUG

/*
	There are 3 different flavors of the following, depending on OBJ_KIND.
*/
#if	OBJ_KIND == OBJ_NO_HEADER

#define obj_alloc(n,class)		malloc(n)
#define obj_free(p)				free(p)
#define obj_new(type,class)		(type *) malloc(sizeof(type))

#elif OBJ_KIND == OBJ_NORMAL

#define obj_alloc(n,class)		m_alloc(n,class)
#define obj_free(p)				m_free(p)
#define obj_new(type,class)		(type *) m_alloc(sizeof(type),class)

#elif OBJ_KIND == OBJ_DEBUG

#define obj_alloc(n,class)		m_alloc_db(n,class)
#define obj_free(p)				m_free_db(p)
#define obj_new(type,class)		(type *) m_alloc_db(sizeof(type),class)

#else
#error "Bad OBJ_KIND"
#endif /* OBJ_KIND cases. */

/*
	There are three definitions of the following routines, which
	do not exist if OBJ_KIND == OBJ_NO_HEADER
*/
#if	OBJ_KIND == OBJ_NO_HEADER

#define obj_check(obj,class,tag)
#define obj_check2(obj,class1,class2,tag)

#elif OBJ_KIND == OBJ_NORMAL

#define obj_check(obj,class,tag)			m_check(obj,class)
#define obj_check2(obj,class1,class2,tag)	m_check2(obj,class1,class2)

#elif OBJ_KIND == OBJ_DEBUG

#define obj_check(obj,class,tag)			m_check_db(obj,class,tag)
#define obj_check2(obj,class1,class2,tag)	m_check2_db(obj,class1,class2,tag)

#else
#error "Bad OBJ_KIND"
#endif /* OBJ_KIND cases. */

/*
	The following are defined unless OBJ_KIND == OBJ_NO_HEADER.
*/
#if OBJ_KIND == OBJ_NO_HEADER

#define obj_type(obj)			OBJ_BAD
#define obj_tag(obj)			""
#define obj_tagType(class)		""

#else

#define obj_type(obj)			m_type(obj)
#define obj_tag(obj)			m_tag(obj)
#define obj_tagType(class)		m_tagType(class)

#endif

/*
	The following are defined only if OBJ_KIND == OBJ_DEBUG.
*/
#if OBJ_KIND == OBJ_DEBUG

#define obj_checkMem(tag)		m_checkMem(tag)
#define obj_checkObj(obj,tag)	m_checkObj(obj,tag)
#define obj_show(flag, tag)		m_show(flag, tag)
#define obj_stat(tag)			m_stat(tag)

#else

#define obj_checkMem(tag)					
#define obj_checkObj(obj,tag)
#define obj_show(flag, tag)
#define obj_stat(tag)

#endif /* OBJ_KIND == OBJ_DEBUG */

/*
	Function prototypes.
	
	Do not call routines starting with m_ directly--call them via the
	corresponding macros instead.
*/
void *		m_alloc_db		(int n, int class);
void *		m_alloc			(int n, int class);
void		m_check			(void * obj, int class);
void		m_check_db		(void * obj, int class, char *tag);
void		m_check2		(void * obj, int class1, int class2);
void		m_check2_db		(void * obj, int class1, int class2, char *tag);
void		m_checkMem		(char * tag);
void		m_checkObj		(void * obj, char * tag);
void		m_free_db		(void * obj);
void		m_free			(void * obj);
void		m_show			(bool all_flag, char *tag);
void		m_stat			(char * tag);
char *		m_tag			(void * obj);
char *		m_tagType		(int class);
int			m_type			(void * obj);
void		obj_test		(void);

#endif /* SPPobj_h_ */
