/*
	SPP: extra string handling library.

	source: SPPstr.c
	started: October 6, 1985
	version: April 6, 1996.
*/

#include "SPP.h"

#include "objlib.h"

#include "SPPerr.h"
#include "SPPobj.h"
#include "SPPpr.h"
#include "SPPstr.h"

/*
	Allocate memory big enough to hold the string,
	then copy the string to the allocated memory.
	
	Warning: 	do NOT use strncpy here, because strncpy
				does not guarantee to give a trailing zero byte
				if the string will not fit!
*/
char *
str_alloc(reg char *s, int class)
{
	FTAG("str_alloc");

	reg char *	p;
	reg int		n;

	n = strlen(s) + 1;
	p = obj_alloc(n, (class | M_STRING | M_NODUMP) );
	strcpy(p, s);

	TRACEP(ftag, eret(); es_str(p); enl());
	
	return p;
}

/*
	Allocate global memory for two strings.
	
	Warning: 	do NOT use strncpy here, because strncpy
				does not guarantee to give a trailing zero byte
				if the string will not fit!
*/
char *
str_2alloc(reg char *s1, reg char *s2, int class)
{
	FTAG("str_2alloc");

	reg char *	p;
	reg int		n1, n2;

	n1 = strlen(s1);
	n2 = strlen(s2);
	p = obj_alloc(n1 + n2 + 1, (class | M_STRING | M_NODUMP) );
	strcpy(p, s1);
	strcpy(p + n1, s2);
	
	TRACEP(ftag, eret(); es_str(p); enl());

	return p;
}

/*
	Concatenate s2 to the end of s1.
	Abort if s2 + s1 is more than n characters.
	
	Warning: 	do NOT use strncpy here, because strncpy
				does not guarantee to give a trailing zero byte
				if the string will not fit!
*/
void
str_ncat(reg char *s1, reg char *s2, int length)
{
	int len1 = strlen(s1);
	int len2 = strlen(s2);
	ASSERT(s1 && s2);
	
	if (len1 + len2 + 1 > length) {
		fatal(es("str_ncat: buffer overflow"); enl());
	}
	
	strcpy(s1 + len1, s2);
}

/*
	Copy s2 to s1.
	Abort if s1 is not large enough to hold s2.
	
	This is *different* from strncpy because:
	
	1.  No padding takes place if s2 is smaller than s1.
	2.  A zero byte *is* guaranteed at the end.
	
	Warning: 	do NOT use strncpy here, because strncpy
				does not guarantee to give a trailing zero byte
				if the string will not fit!
*/
void
str_ncpy(reg char *s1, reg char *s2, int length)
{
	ASSERT(s1 && s2);

	if (strlen(s2) + 1 > length) {
		fatal(es("str_ncpy: buffer overflow"); enl());
	}

	strcpy(s1,s2);
}
